#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/pkcs12.h>
#include <openssl/opensslv.h>
#include <openssl/bn.h>
#include <openssl/err.h>

#include <typeinfo>

#include "datetime.h"
#include "dcisec.h"
#include "logger.h"
#include "opensslutil.h"

#define OpenSSLError() \
{ \
        unsigned long errors; \
        char errors_buf[256] = { 0 }; \
        do{ \
                errors = ERR_get_error(); \
                if (errors != 0) { \
                        CredentialLogger.msg(ERROR, "OpenSSL error string: %s", ERR_error_string(errors, errors_buf)); \
                } \
        }  while(errors != 0); \
}

namespace AuthN {

using namespace AuthN::Utils;
using namespace AuthN::OpenSSL;

  Logger& CredentialLogger = log();

  static std::string tostring(unsigned int val) {
    std::stringstream ss;
    ss << val;
    return ss.str();
  }

  Credentials::Credentials(const AuthN::Context& ctx):valid_(false),priv_key_(NULL),cert_(NULL),cert_chain_(NULL), validator_(NULL), proxy_policy_(NULL) {
    try {
    context_ = &ctx.Copy();
    std::string certfile = ctx.GetCertPath();
    std::string keyfile = ctx.GetKeyPath();
    std::string passphrase4key; // = ctx.PasswordCallback();
    if(!certfile.empty()) loadCertificateFile(certfile, cert_, &cert_chain_);
    if(!keyfile.empty()) loadKeyFile(keyfile, priv_key_, passphrase4key);
    else if(!certfile.empty()) loadKeyFile(certfile, priv_key_, passphrase4key); // in this case, the key is supposed to be inside the proxy file
    valid_ = true;
    } catch (Status& err) {
      last_error_ = err;
    };
  };


  Credentials::~Credentials(void) {
    if(cert_) X509_free(cert_);
    if(cert_chain_) sk_X509_pop_free(cert_chain_, X509_free);
    if(priv_key_) EVP_PKEY_free(priv_key_);
    if(validator_) delete validator_;
    if(context_) delete context_;
    if(proxy_policy_) delete proxy_policy_;
  }

  Credentials& Credentials::Copy(void) const {
    Credentials* ncred = new Credentials(Context());
    ncred->Assign(cert_,cert_chain_,priv_key_);
    if(context_) ncred->SetContext(*context_);
    if(validator_) ncred->SetValidator(*validator_); 
    return *ncred;
  }

  AuthN::Status Credentials::Assign(X509 *cert, const STACK_OF(X509)* chain, const EVP_PKEY *key) {
    if(cert) {
      if(cert_) X509_free(cert_);
      cert_ = X509_dup(cert);
    }
    else CredentialLogger.msg(INFO,"Certificate is empty");

    if(chain && sk_X509_num(chain) != 0) {
      if(cert_chain_) sk_X509_pop_free(cert_chain_, X509_free);
      cert_chain_ = sk_X509_new_null();
      for (int i=0; i < sk_X509_num(chain); i++) {
        X509* tmp = X509_dup(sk_X509_value(chain,i));
        sk_X509_insert(cert_chain_, tmp, i);
      }
    }
    else CredentialLogger.msg(INFO,"Certificate chain is empty");

    if(key) {
      if(priv_key_ != NULL) EVP_PKEY_free(priv_key_);
      priv_key_ = EVP_PKEY_new();
      RSA* priv_rsa = NULL;
      if(key->type == EVP_PKEY_RSA && key->pkey.rsa != NULL)
        priv_rsa = key->pkey.rsa;
      EVP_PKEY_set1_RSA(priv_key_, priv_rsa);
    }
    else CredentialLogger.msg(INFO,"Private key is empty");

    valid_ = true;

    return Status(0);
  }
  
  AuthN::Status Credentials::Assign(std::istream& cert, std::istream& chain, std::istream& key) {
    std::string cert_str;
    std::string chain_str;
    std::string key_str;

    std::getline<char>(cert, cert_str, 0);
    if (!cert) return Status(-1);

    std::getline<char>(chain, chain_str, 0);
    if (!chain) return Status(-1);

    std::getline<char>(key, key_str, 0);
    if (!key) return Status(-1);

    return(Assign(cert_str, chain_str, key_str));
  }

  AuthN::Status Credentials::Assign(const std::string& cert, const std::string& chain, const std::string& key) {
    try {

    AutoBIO certbio(BIO_new_mem_buf((void*)(cert.c_str()), cert.length()));
    if(!certbio) return Status(-1,"Can not read certificate string: "+GetOpenSSLError());
    loadCertificate(certbio, cert_, &cert_chain_, false);

    AutoBIO chainbio(BIO_new_mem_buf((void*)(chain.c_str()), chain.length()));
    if(!chainbio) return Status(-1,"Can not read certificate chain string: "+GetOpenSSLError());
    loadCertificateChain(chainbio, &cert_chain_);

    AutoBIO key_bio(BIO_new_mem_buf((unsigned char*)(key.c_str()), key.length()));
    unsigned char *passphrase = NULL; //TODO: input passphrase
    priv_key_ = PEM_read_bio_PrivateKey(key_bio, NULL, 0, passphrase);

    valid_ = true;

    return Status(0);

    } catch (Status& err) {
      last_error_ = err;
    }

    return last_error_;
  }

  X509* Credentials::GetCertificate(void) const {
    return cert_;
  }

  void Credentials::GetCertificate(std::string& str) const {
    if(!cert_) return;
    BIO *out = BIO_new(BIO_s_mem());
    if(!out) return;
    if(!PEM_write_bio_X509(out,cert_)) { BIO_free_all(out); return; };

    for(;;) {
      char s[256];
      int l = BIO_read(out,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(out);
    return;
  }

  void Credentials::GetCertificate(std::ostream& o) const {
    std::string s;
    GetCertificate(s);
    o << s;
  }

  STACK_OF(X509)* Credentials::GetChain(void) const {
    return cert_chain_;
  }

  void Credentials::GetChain(std::string& str) const {
    BIO *out = BIO_new(BIO_s_mem());
    if(!out) return;
    X509 *cert;
    if(cert_chain_) CredentialLogger.msg(DEBUG, "Certiticate chain number %d",sk_X509_num(cert_chain_));

    if(cert_chain_) for (int n = 0; n < sk_X509_num(cert_chain_); n++) {
      cert = sk_X509_value(cert_chain_, n);
      if(!PEM_write_bio_X509(out,cert)) { BIO_free_all(out); return; };
      for(;;) {
        char s[256];
        int l = BIO_read(out,s,sizeof(s));
        if(l <= 0) break;
        str.append(s,l);
      }
    }
    BIO_free_all(out);
    return;
  }

  void Credentials::GetChain(std::ostream& o) const {
    std::string s;
    GetCertificate(s);
    o << s;
  }

  EVP_PKEY* Credentials::GetPrivateKey(void) const {
    return priv_key_; 
  }
      
  void Credentials::GetPrivateKey(std::string& str, bool encrypt) const {
    if(!priv_key_) return;

    BIO* priv_bio = BIO_new(BIO_s_mem());
    const EVP_CIPHER *cipher = NULL;
    if(encrypt) cipher = EVP_des_ede3_cbc();
    if(!encrypt) {
      if(!PEM_write_bio_PrivateKey(priv_bio, priv_key_, cipher,
            NULL, 0, NULL, NULL)) { BIO_free_all(priv_bio); return; }
    }
    else {
      PW_CB_DATA cb_data;
      cb_data.password = NULL;
      cb_data.prompt_info = NULL;
      if(!PEM_write_bio_PrivateKey(priv_bio,priv_key_,cipher,NULL,0, (pem_password_cb *)passwordcb,&cb_data)) {
        BIO_free_all(priv_bio); return;
      }
    }

    for(;;) {
      char s[256];
      int l = BIO_read(priv_bio,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(priv_bio);
  }
      
  void Credentials::GetPrivateKey(std::ostream& o, bool encrypt) const {
    std::string s;
    GetPrivateKey(s,encrypt);
    o << s;
  }

  AuthN::Status Credentials::Validate(void) {
    if(validator_ == NULL) {
      last_error_ = AuthN::Status(-1,"No validator is assigned");
    } else {
      last_error_ = validator_->Validate(*this);
    };
    return last_error_;
  }

#if 0
#define SERIAL_RAND_BITS        64
  static int rand_serial(ASN1_INTEGER *ai) {
    int ret = 0;
    BIGNUM *btmp = BN_new();
    if(!btmp) goto error;
    if(!BN_pseudo_rand(btmp, SERIAL_RAND_BITS, 0, 0)) goto error;
    if(ai && !BN_to_ASN1_INTEGER(btmp, ai)) goto error;
    ret = 1;
  error:
    if(btmp) BN_free(btmp);
    return ret;
  }
#endif

  AuthN::Status Credentials::Sign(CredentialsRequest& request, Credentials& out, const std::string& conf_file) {
    const X509_REQ* req = request.GetRequest();
    std::string dn_name = request.GetSubjectName();
    if(req == NULL) {
      last_error_ = Status(-1,"The X509 request is empty");
      return last_error_;
    }

    if(cert_ == NULL ) {
      //If the X509 object is empty, we look the signing as self-signing, 
      //for an EEC certificate signing, self-signing only happens if the EEC is a CA certificate
      X509* cert = NULL;
      time_t from = request.GetValidFrom();
      time_t till = request.GetValidTill();
      const EVP_PKEY* privkey = request.GetPrivateKey();
      bool ret = selfsign_request(&cert, req, privkey, from, till, dn_name, -1, conf_file);
      if(!ret) {
        last_error_ = Status(-1,"Failed to sign a CA certificate");
        return last_error_;
      };
      Assign(cert, NULL, privkey);
      out.Assign(cert, NULL, privkey);
      try {
        CACredentialsRequest& carequest = dynamic_cast<CACredentialsRequest&>(request);
        carequest.Assign(cert);
      } catch(const std::bad_cast& e) {
        last_error_ = Status(-1,"Credentials request is not CA request");
        if(cert != NULL) X509_free(cert);
        return last_error_;
      }
      if(cert != NULL) X509_free(cert);
      return Status(0);
    }
    std::string ext_name = "basicConstraints";
    Credentials::Extension ext;
    GetExtension(ext_name, ext);
    if(ext.value.find("CA:TRUE") != std::string::npos) {
      //If the basicConstraints extension is CA:TRUE, then an EEC certificate will be signed
      X509* cert = NULL;
      time_t from = request.GetValidFrom();
      time_t till = request.GetValidTill();
      std::string serial_file = ""; //TODO
      bool ret = sign_request(&cert, req, priv_key_, cert_, from, till, dn_name, serial_file, conf_file);
      if(!ret) {
        last_error_ = Status(-1,"Failed to sign an EEC certificate");
        return last_error_;
      }
      const EVP_PKEY* privkey = request.GetPrivateKey();

      STACK_OF(X509)* chain = NULL;
      chain = sk_X509_new_null();
      X509* tmp = X509_dup(cert_);
      sk_X509_insert(chain, tmp, 0);
      out.Assign(cert, chain, privkey);

      if(chain) sk_X509_pop_free(chain, X509_free);
      if(cert != NULL) X509_free(cert);
      return Status(0);
    }

    X509* cert = NULL;
    time_t from = request.GetValidFrom();
    time_t till = request.GetValidTill();

    std::string proxypolicy;
/*
    Credentials::Extension policy_ext;
    try {
      ProxyCredentialsRequest& proxyrequest = dynamic_cast<ProxyCredentialsRequest&>(request);
      proxyrequest.GetPolicy(policy_ext);
    } catch(const std::bad_cast& e) {
      CredentialLogger.msg(ERROR, "Failed to cast CredentialsRequest to ProxyCredentialsRequest: %s", e.what());
    }
    proxypolicy=policy_ext.value;
*/

    bool ret = sign_proxy_request(&cert, req, priv_key_, cert_, from, till, proxypolicy, conf_file);
    if(!ret) {
      last_error_ = Status(-1,"Failed to sign a proxy certificate");
      return last_error_;
    }
    const EVP_PKEY* privkey = request.GetPrivateKey();
    STACK_OF(X509)* chain = NULL;
    chain = sk_X509_new_null();
    X509* tmp = X509_dup(cert_);
    sk_X509_insert(chain, tmp, 0);
    if(cert_chain_ && sk_X509_num(cert_chain_) != 0) {
      for(int i=0; i < sk_X509_num(cert_chain_); i++) {
        X509* tmp = X509_dup(sk_X509_value(cert_chain_,i));
        sk_X509_insert(chain, tmp, i+1);
      }
    }

    last_error_ = out.Assign(cert, chain, privkey);
    if(cert != NULL) X509_free(cert);
    if(chain) sk_X509_pop_free(chain, X509_free);
    return last_error_;
  }

  std::string Credentials::GetSubjectName(void) const {
    X509_NAME *subject = NULL;
    if(!cert_) return "";
    subject = X509_get_subject_name(cert_);
    std::string str;
    char buf[256];
    if(subject!=NULL)
      X509_NAME_oneline(subject,buf,sizeof(buf));
    str.append(buf);
    return str;
  }

  std::string Credentials::GetIdentityName(void) const {
    // TODO: it is more correct to go through chain till first non-proxy cert
    // TODO: different format must be implemented
    X509_NAME *subject = NULL;
    X509_NAME_ENTRY *ne = NULL;
    if(!cert_) return "";
    subject = X509_NAME_dup(X509_get_subject_name(cert_));

#if 0
    int proxy_depth = verify_ctx_.proxy_depth;
    for(int i=0; i<proxy_depth; i++) {
      ne = X509_NAME_delete_entry(subject, X509_NAME_entry_count(subject)-1);
      if(ne)
        X509_NAME_ENTRY_free(ne);
    }
#endif

    ASN1_STRING* entry;
    std::string entry_str;
    for(;;) {
      ne = X509_NAME_get_entry(subject, X509_NAME_entry_count(subject)-1);
      if (!OBJ_cmp(ne->object,OBJ_nid2obj(NID_commonName))) {
        entry = X509_NAME_ENTRY_get_data(ne);
        entry_str.assign((const char*)(entry->data), (std::size_t)(entry->length));
        if(entry_str == "proxy" || entry_str == "limited proxy" ||
           entry_str.find_first_not_of("0123456789") == std::string::npos) {
         //Drop the name entry "proxy", "limited proxy", or the random digital(RFC)
          ne = X509_NAME_delete_entry(subject, X509_NAME_entry_count(subject)-1);
          X509_NAME_ENTRY_free(ne);
          ne = NULL;
        }
        else break;
      }
      else break;
    }
    char buf[256];
    if(subject!=NULL) {
      X509_NAME_oneline(subject,buf,sizeof(buf)-1);
      X509_NAME_free(subject);
    }
    buf[sizeof(buf)-1] = 0;
    std::string str(buf);
    return str;
  }

  std::string Credentials::GetIssuerName(int position) const {
    X509_NAME *issuer = NULL;
    std::string str;
    if(position == 0) { 
      if(!cert_ ) return "";
      issuer = X509_get_issuer_name(cert_);
      char buf[256];
      if(issuer!=NULL)
        X509_NAME_oneline(issuer,buf,sizeof(buf));
      str.append(buf);
    }
    else if (cert_chain_ && sk_X509_num(cert_chain_) != 0) {
      for (int i=0; i < sk_X509_num(cert_chain_); i++) {
        if(position - 1 == i) {
          issuer = X509_get_issuer_name(sk_X509_value(cert_chain_,i));
          char buf[256];
          if(issuer!=NULL)
            X509_NAME_oneline(issuer,buf,sizeof(buf));
          str.append(buf);
        }
      }
    }
    return str;
  }

  AuthN::Status Credentials::GetChainCert(Credentials& cert, int position) const {
    Status status;
    if(position == 0) {
      if(cert_ != NULL) { 
        return cert.Assign(cert_, NULL, NULL);
      }
      else { 
        return Status(-1,"The main certificate is empty"); 
      }
    }
    if(cert_chain_ && sk_X509_num(cert_chain_) != 0) {
      for (int i=0; i < sk_X509_num(cert_chain_); i++) {
        if(position-1 == i) {
          return cert.Assign(sk_X509_value(cert_chain_,i), NULL, NULL);
        }
      }
    }
    return Status(-1,"Failed to get "+tostring(position)+" certificate in chain");
  }

  time_t Credentials::GetValidFrom(void) {
    X509* tmp_cert = NULL;
    Time start_time(-1);
    int n;
    ASN1_UTCTIME* atime = NULL;
    if(cert_ == NULL) {
      return -1;
    }
    if(cert_chain_) for (n = 0; n < sk_X509_num(cert_chain_); n++) {
      tmp_cert = sk_X509_value(cert_chain_, n);
      atime = X509_get_notBefore(tmp_cert);
      Time s = asn1_to_utctime(atime);
      if (start_time == Time(-1) || s > start_time) { start_time = s; }
    }
    atime = X509_get_notBefore(cert_);
    Time s = asn1_to_utctime(atime);
    if (start_time == Time(-1) || s > start_time) { start_time = s; }
    return start_time.GetTime();
  }

  time_t Credentials::GetValidTill(void) {
    X509* tmp_cert = NULL;
    Time end_time(-1);
    int n;
    ASN1_UTCTIME* atime = NULL;
    if(cert_ == NULL) {
      return -1;
    }
    if(cert_chain_) for (n = 0; n < sk_X509_num(cert_chain_); n++) {
      tmp_cert = sk_X509_value(cert_chain_, n);
      atime = X509_get_notAfter(tmp_cert);
      Time e = asn1_to_utctime(atime);
      if (end_time == Time(-1) || e < end_time) { end_time = e; }
    }
    atime = X509_get_notAfter(cert_);
    Time e = asn1_to_utctime(atime);
    if (end_time == Time(-1) || e < end_time) { end_time = e; }
    return end_time.GetTime();
  }

  bool Credentials::GetSelfSigned(void) {
    std::string sub, iss;
    sub = GetSubjectName();
    iss = GetIssuerName();
    if(!iss.empty() && (iss == sub)) return true;
    else return false;
  }

  bool Credentials::GetCA(void) {
    int pos = X509_get_ext_by_NID(cert_, NID_basic_constraints, -1);
    bool ret = false;
    if(pos != -1) {
      X509_EXTENSION *ex = X509_get_ext(cert_, pos);
      if(ex) {
        BASIC_CONSTRAINTS *bs = (BASIC_CONSTRAINTS *)X509V3_EXT_d2i(ex);
        ret = (bs->ca ? true: false);
        BASIC_CONSTRAINTS_free(bs);
      }
    }
    return ret;
  }

  bool Credentials::GetProxy(void) {
    bool ret = false;
    int pos = X509_get_ext_by_NID(cert_, NID_proxyCertInfo, -1);
    if(pos != -1) ret = true;
    return ret;
  }

  bool Credentials::GetProxyLimited(void) {
    std::string sub_name = (*this).GetSubjectName();
    if(sub_name.find("cn=limited proxy") != std::string::npos) return true;
    else return false;
  }

  bool Credentials::GetProxyPolicy(Extension& policy) {
    if(proxy_policy_ != NULL) { 
      //If the proxy has been validated, then the proxy policy should have already been parsed.
      policy = *proxy_policy_; return true; 
    }
    else {
      if(cert_ == NULL) return false;
      int pos = X509_get_ext_by_NID(cert_, NID_proxyCertInfo, -1);
      X509_EXTENSION* ex = X509_get_ext(cert_, pos);
      if(ex) {
        PROXY_CERT_INFO_EXTENSION *pci = (PROXY_CERT_INFO_EXTENSION *)X509V3_EXT_d2i(ex);
        std::string oid;
        char buf[256];
        int buflen;
        buflen = OBJ_obj2txt(buf, sizeof(buf), pci->proxyPolicy->policyLanguage, 0);
        if(buflen < 0) { CredentialLogger.msg(ERROR,"Failed to call OBJ_obj2txt to convert obj to txt");}
        else oid = buf;
        policy.oid = oid;
        if(X509_EXTENSION_get_critical(ex)) policy.critical = true;
        else policy.critical = false;
        policy.value = std::string((const char*)(pci->proxyPolicy->policy->data), pci->proxyPolicy->policy->length);
      } 
      return true;
    }
    return false;
  }

  static std::string get_obj_txt(X509_EXTENSION* ext) {
    ASN1_OBJECT* obj;
    obj = X509_EXTENSION_get_object(ext);
    char buf[256];
    int buflen;
    buflen = OBJ_obj2txt(buf, sizeof(buf), obj, 0);
    std::string str = std::string(buf, buflen);
    return str;
  }

  bool Credentials::GetExtension(int pos, Extension& ext) {
    std::string oid;
    bool crit = false;
    std::string ext_val;
    if(cert_ == NULL) return false;
    int num = X509_get_ext_count(cert_);
    if ((num > 0) && (pos < num)) {
      X509_EXTENSION* x509_ext = NULL;
      x509_ext = X509_get_ext(cert_, pos);
      if(!parse_extension(x509_ext, crit, oid, ext_val)) return false;
    }
    else if(pos >=num) {
      CredentialLogger.msg(ERROR,"The ext position %d exceeds the ext number %d", pos, num);
      return false;
    }
    ext.critical = crit;
    ext.oid = oid;
    ext.value = ext_val;
    return true;
  }

  bool Credentials::GetExtension(const std::string& name, Extension& ext) {
    std::string oid;
    bool crit = false;
    std::string ext_val;
    if(cert_ == NULL) return false;
    int nid;
    nid = OBJ_txt2nid(name.c_str());
    if(nid == NID_undef) nid = OBJ_sn2nid(name.c_str());
    if(nid == NID_undef) nid = OBJ_ln2nid(name.c_str());
    if(nid != NID_undef) {
      int pos = X509_get_ext_by_NID(cert_, nid, -1);
      if (pos != -1) {
        X509_EXTENSION* x509_ext = NULL;
        x509_ext = X509_get_ext(cert_, pos);
        if(!parse_extension(x509_ext, crit, oid, ext_val)) return false;
      }
      ext.critical = crit;
      ext.oid = oid;
      ext.value = ext_val;
      return true;
    }
    else { // If the 'name' argument is an oid, and this oid is not registered in openssl
      int num = X509_get_ext_count(cert_);
      if (num > 0) {
        for(int i = 0; i < num; i++) {
          X509_EXTENSION* x509_ext = NULL;
          x509_ext = X509_get_ext(cert_, i);
          std::string str = get_obj_txt(x509_ext);
          if(str == name) {
            if(!parse_extension(x509_ext, crit, oid, ext_val)) return false;
            ext.critical = crit;
            ext.oid = oid;
            ext.value = ext_val;
            return true;
          }
        }
        return false;
      }
    }
    return false;
  }

  bool Credentials::GetAttributes(const std::string& name, std::list<std::string>& attrs) {
    //1. Suppose the attribute is a extension, which means the name argument can be matched into an extension name
    bool crit = false;
    std::string ext_val;
    if(cert_ == NULL) return false;
    int nid;
    nid = OBJ_txt2nid(name.c_str());
    if(nid == NID_undef) nid = OBJ_sn2nid(name.c_str());
    if(nid == NID_undef) nid = OBJ_ln2nid(name.c_str());
    if(nid != NID_undef) {
      int pos = X509_get_ext_by_NID(cert_, nid, -1);
      if (pos != -1) {
        X509_EXTENSION* x509_ext = NULL;
        x509_ext = X509_get_ext(cert_, pos);
        std::string attr;
        attr.append((char*)(x509_ext->value->data), x509_ext->value->length);
        attrs.push_back(attr);
        return true;
      }
    }

    return false; //TODO
  }

  void Credentials::SetContext(const AuthN::Context& env) {
    context_ = &env.Copy();
  }

  const AuthN::Context& Credentials::GetContext(void) const {
    return *context_;
  }

  void Credentials::SetValidator(const AuthN::Validator& validator) {
    if(validator_) delete validator_;
    validator_ = &(validator.Copy());
  }

  const AuthN::Validator& Credentials::GetValidator(void) const {
    return *validator_;
  }


}
