
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/opensslv.h>
#include <openssl/bn.h>
#include <openssl/err.h>

#include "opensslutil.h"
#include "dcisec.h"
#include "logger.h"

#define OpenSSLErrors() \
{ \
        unsigned long errors; \
        char errors_buf[256] = { 0 }; \
        do{ \
                errors = ERR_get_error(); \
                if (errors != 0) { \
                        std::cout<<"OpenSSL gives error: "<<ERR_error_string(errors, errors_buf)<<std::endl; \
                } \
        }  while(errors != 0); \
}

namespace AuthN {

using namespace AuthN::OpenSSL;

  Logger& CredentialRequestLogger = log();

  CredentialsRequest::CredentialsRequest(const AuthN::Context& ctx) :
      x509_req_(NULL), pub_key_(NULL), priv_key_(NULL) {
  }

  CredentialsRequest::~CredentialsRequest() {
    if(x509_req_) X509_REQ_free(x509_req_);
    if(priv_key_) EVP_PKEY_free(priv_key_);
    if(pub_key_) EVP_PKEY_free(pub_key_);
  }

  CredentialsRequest& CredentialsRequest::Copy(void) const {
    CredentialsRequest* req = new CredentialsRequest(Context());
    req->AssignRequest(x509_req_, priv_key_);
    return *req;    
  }

  static int keygen_cb(int p, int, BN_GENCB*) {
    char c='*';
    if (p == 0) c='.';
    if (p == 1) c='+';
    if (p == 2) c='*';
    if (p == 3) c='\n';
    std::cerr<<c;
    return 1;
  }

  AuthN::Status CredentialsRequest::MakeKeys(int bits) {
    bool res = false;
    RSA* rsa_key = NULL;
    const EVP_MD *digest = (EVP_MD*)EVP_sha1();;
    EVP_PKEY* pkey;
    int keybits = bits;

#ifdef HAVE_OPENSSL_OLDRSA
    unsigned long prime = RSA_F4;
    rsa_key = RSA_generate_key(keybits, prime, keygen_cb, NULL);
    if(!rsa_key) {
      std::cout<<"RSA_generate_key failed"<<std::endl;
      if(rsa_key) RSA_free(rsa_key);
      return Status(-1);
    }
#else
    BN_GENCB cb;
    BIGNUM *prime = BN_new();
    if(!prime){ std::cout<<"BN_new fails"<<std::endl; return Status(-1); }

    rsa_key = RSA_new();
    if(!rsa_key) { std::cout<<"RSA_new fails"<<std::endl; BN_free(prime); return Status(-1); }
    BN_GENCB_set(&cb,&keygen_cb,NULL);
    int ret = BN_set_word(prime,RSA_F4);
    if(ret != 1) {
      BN_free(prime); RSA_free(rsa_key);
      return Status(-1);
    }
    ret = RSA_generate_key_ex(rsa_key, keybits, prime, &cb);
    if(ret != 1) {
      std::cout<<"RSA_generate_key_ex failed"<<std::endl;
      BN_free(prime); RSA_free(rsa_key);
      return Status(-1);
    }
#endif
    RSA* priv = NULL; RSA* pub = NULL;
    priv = RSAPrivateKey_dup(rsa_key);
    if (!priv) { RSA_free(rsa_key); return Status(-1); }

    pub = RSAPublicKey_dup(rsa_key);
    if (!pub) { 
      RSA_free(priv); RSA_free(rsa_key);
      return Status(-1);
    }
    RSA_free(rsa_key);

    pub_key_ = EVP_PKEY_new();
    priv_key_ = EVP_PKEY_new();
    EVP_PKEY_set1_RSA(pub_key_, pub);
    EVP_PKEY_set1_RSA(priv_key_, priv);

    RSA_free(priv); priv = NULL;
    RSA_free(pub); pub = NULL;

    return Status(0);
  }

  const EVP_PKEY* CredentialsRequest::GetPrivateKey(void) const{
    return priv_key_;
  }

  void CredentialsRequest::GetPrivateKey(std::string& str, bool encrypt) const{
    if(!priv_key_) CredentialRequestLogger.msg(ERROR, "Failed to get private key");

    BIO* priv_bio = BIO_new(BIO_s_mem());
    const EVP_CIPHER *cipher = NULL;
    if(encrypt) cipher = EVP_des_ede3_cbc();
    if(!encrypt) {
      if(!PEM_write_bio_PrivateKey(priv_bio, priv_key_, cipher,
            NULL, 0, NULL, NULL)) { BIO_free_all(priv_bio); return; }
    }
    else {
      PW_CB_DATA cb_data;
      cb_data.password = NULL;
      cb_data.prompt_info = NULL;
      if(!PEM_write_bio_PrivateKey(priv_bio,priv_key_,cipher,NULL,0, (pem_password_cb *)passwordcb,&cb_data)) {
        BIO_free_all(priv_bio); return;
      }
    }

    for(;;) {
      char s[256];
      int l = BIO_read(priv_bio,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(priv_bio);
  }

  void CredentialsRequest::GetPrivateKey(std::ostream& o, bool encrypt) const{
    std::string s;
    GetPrivateKey(s,encrypt);
    o << s;
  }

  const EVP_PKEY* CredentialsRequest::GetPublicKey(void) const{
    return pub_key_;
  }

  void CredentialsRequest::GetPublicKey(std::string& str) const{
    BIO* pub_bio = BIO_new(BIO_s_mem());
    PEM_write_bio_PrivateKey(pub_bio, pub_key_, NULL,
            NULL, 0, NULL, NULL); 
    for(;;) {
      char s[256];
      int l = BIO_read(pub_bio,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(pub_bio);
  }

  void CredentialsRequest::GetPublicKey(std::ostream& o) const{
    std::string s;
    GetPrivateKey(s);
    o << s;
  }

  AuthN::Status CredentialsRequest::MakeRequest(const std::string& conf_file) {
    x509_req_ = X509_REQ_new();
    if(!x509_req_) { return Status(-1); }
    if(!make_request(x509_req_, pub_key_, priv_key_, false, conf_file, subject_name_)) { return Status(-1); }
    return Status(0);
//    if(!X509_REQ_set_version(x509_req_,3L)) { return Status(-1); } 
//    X509_NAME *name = NULL;
//    unsigned long chtype = MBSTRING_ASC;  //TODO
//    name = parse_name((char*)(dn.c_str()), chtype, 0);
//    X509_REQ_set_subject_name(req, name);
//    X509_NAME_free(name);

//    if(!X509_REQ_set_pubkey(x509_req_, pub_key_)) { return Status(-1); }
//    const EVP_MD *digest = (EVP_MD*)EVP_sha1();
//    if(!X509_REQ_sign(x509_req_, priv_key_, digest)) { return Status(-1); }
//    return Status(0);
  }

  const X509_REQ* CredentialsRequest::GetRequest(void) const {
    return x509_req_;
  }

  void CredentialsRequest::GetRequest(std::string& str) const {
    BIO* req_bio = BIO_new(BIO_s_mem());
    if(!req_bio) {
      std::cout<<"Can not create BIO for request"<<std::endl;
      OpenSSLErrors(); return;
    }
    if(!PEM_write_bio_X509_REQ(req_bio, x509_req_)){ return; }

    for(;;) {
      char s[256];
      int l = BIO_read(req_bio,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }

    BIO_free_all(req_bio);
    return;
  }

  void CredentialsRequest::GetRequest(std::ostream& o) const {
    std::string s;
    GetRequest(s);
    o << s;
  }

  AuthN::Status CredentialsRequest::AssignKeys(const EVP_PKEY* pub, const EVP_PKEY* priv) {
    if(pub_key_ != NULL) EVP_PKEY_free(pub_key_);
    if(priv_key_ != NULL) EVP_PKEY_free(priv_key_);
    pub_key_ = EVP_PKEY_new();
    priv_key_ = EVP_PKEY_new();
    RSA* pub_rsa = NULL;
    if(pub->type == EVP_PKEY_RSA && pub->pkey.rsa != NULL)
      pub_rsa = pub->pkey.rsa;
    EVP_PKEY_set1_RSA(pub_key_, pub_rsa);
    RSA* priv_rsa = NULL;
    if(priv->type == EVP_PKEY_RSA && priv->pkey.rsa != NULL)
      priv_rsa = priv->pkey.rsa;
    EVP_PKEY_set1_RSA(priv_key_, priv_rsa);
    return Status(0);
  }

  AuthN::Status CredentialsRequest::AssignKeys(std::istream& pub, std::istream& priv) {
    std::string pub_str;
    std::string priv_str;

    std::getline<char>(pub, pub_str, 0);
    if (!pub) return Status(-1);

    std::getline<char>(priv, priv_str, 0);
    if (!priv) return Status(-1);
    return(AssignKeys(pub_str, priv_str));
  }

  AuthN::Status CredentialsRequest::AssignKeys(const std::string& pub, const std::string& priv) {
    const unsigned char* pub_str = (const unsigned char*)(pub.c_str());
    pub_key_ = d2i_PublicKey(EVP_PKEY_RSA, NULL, &pub_str, pub.length());

    BIO* priv_bio;
    priv_bio = BIO_new_mem_buf((unsigned char*)(priv.c_str()), priv.length());
    unsigned char *passphrase = NULL; //TODO: input passphrase
    priv_key_ = PEM_read_bio_PrivateKey(priv_bio, NULL, 0, passphrase);
    BIO_free(priv_bio);

    return Status(0);
  }

  AuthN::Status CredentialsRequest::AssignRequest(X509_REQ* req, const EVP_PKEY* priv) {
    if(req!=NULL) { 
      if(x509_req_ != NULL) X509_REQ_free(x509_req_);
      x509_req_ = X509_REQ_dup(req);
    }
    else { std::cout<<"x509_req_ is not valid"; return Status(-1); }

    if(priv_key_ != NULL) EVP_PKEY_free(priv_key_);
    priv_key_ = EVP_PKEY_new();
    RSA* priv_rsa = NULL;
    if(priv->type == EVP_PKEY_RSA && priv->pkey.rsa != NULL)
      priv_rsa = priv->pkey.rsa;
    EVP_PKEY_set1_RSA(priv_key_, priv_rsa);

    return Status(0);
  }

  AuthN::Status CredentialsRequest::AssignRequest(std::istream& req, std::istream& priv) {
    std::string req_str;
    std::string priv_str;

    std::getline<char>(req, req_str, 0);
    if (!req) return Status(-1);

    std::getline<char>(priv, priv_str, 0);
    if (!priv) return Status(-1);
    return(AssignRequest(req_str, priv_str));
  }

  AuthN::Status CredentialsRequest::AssignRequest(const std::string& req, const std::string& priv) {
    BIO* req_bio;
    req_bio = BIO_new_mem_buf((unsigned char*)(req.c_str()), req.length());
    x509_req_ = PEM_read_bio_X509_REQ(req_bio,NULL,NULL,NULL);//d2i_X509_REQ_bio(req_bio, NULL);
    BIO_free(req_bio);

    BIO* priv_bio;
    priv_bio = BIO_new_mem_buf((unsigned char*)(priv.c_str()), priv.length());
    unsigned char *passphrase = NULL; //TODO: input passphrase
    priv_key_ = PEM_read_bio_PrivateKey(priv_bio, NULL, 0, passphrase);
    BIO_free(priv_bio);

    return Status(0);
  }

  AuthN::Status CredentialsRequest::SetSubjectName(const std::string sn) {
    subject_name_ = sn;
    //TODO: parse the sn, and compose it into RFC2253
    return Status(0);
  }

  std::string CredentialsRequest::GetSubjectName(void) const {
    return subject_name_;
  }

  time_t CredentialsRequest::GetValidFrom(void) const {
    return valid_from_;
  }

  void CredentialsRequest::SetValidFrom(time_t sec) {
    valid_from_ = sec;
  }

  time_t CredentialsRequest::GetValidTill(void) const {
    return valid_till_;
  }

  void CredentialsRequest::SetValidTill(time_t sec) {
    valid_till_ = sec;
  }

  bool CredentialsRequest::GetExtension(int pos, Credentials::Extension& ext) {
    std::string oid;
    bool crit = false;
    std::string ext_val;

    STACK_OF(X509_EXTENSION)* exts = NULL;
    X509_EXTENSION* x509_ext;
    exts = X509_REQ_get_extensions(x509_req_);
    int num = sk_X509_EXTENSION_num(exts);
    if((num > 0) && (pos < num)) {
      x509_ext = sk_X509_EXTENSION_value(exts, pos);
      if(!parse_extension(x509_ext, crit, oid, ext_val)) return false;
    }
    else if(pos >=num) { CredentialRequestLogger.msg(ERROR,"The ext position %d exceeds the ext number %d", pos, num); return false; }

    ext.critical = crit;
    ext.oid = oid;
    ext.value = ext_val;
    return true;
  }

  bool CredentialsRequest::GetExtension(const std::string& name, Credentials::Extension& ext) {
    std::string oid;
    bool crit = false;
    std::string ext_val;

    STACK_OF(X509_EXTENSION)* exts = NULL;
    X509_EXTENSION* x509_ext;
    exts = X509_REQ_get_extensions(x509_req_);
    int num = sk_X509_EXTENSION_num(exts);
    for(int i = 0; i < num; i++) {
      const char *extname;
      x509_ext = sk_X509_EXTENSION_value(exts, i);
      extname = OBJ_nid2sn(OBJ_obj2nid(X509_EXTENSION_get_object(x509_ext)));
      if (strcmp(extname, name.c_str()) == 0) {
        if(!parse_extension(x509_ext, crit, oid, ext_val)) return false;
      }
    }
    ext.critical = crit;
    ext.oid = oid;
    ext.value = ext_val;
    return true;
  }

  bool CredentialsRequest::AddExtension(Credentials::Extension& ext) {
    X509_EXTENSION* x509_ext = NULL;
    STACK_OF(X509_EXTENSION)* exts = NULL;
    exts = X509_REQ_get_extensions(x509_req_);
    if(ext.oid.empty()) return false;
    x509_ext = X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid((char*)(ext.oid.c_str())), (char*)(ext.value.c_str()));
    if(x509_ext && sk_X509_EXTENSION_push(exts, x509_ext)) return true;
    else return false;
  }

  bool CredentialsRequest::GetAttributes(const std::string& name, std::list<std::string>& attrs) {
    return false;
  }

  bool CredentialsRequest::AddAttributes(const std::string& name, std::list<std::string>& attrs) {
    return false;
  }

  CACredentialsRequest::CACredentialsRequest(const AuthN::Context& ctx) : CredentialsRequest(ctx), cert_(NULL) {
  }

  CACredentialsRequest::~CACredentialsRequest(void) {
    if(cert_) X509_free(cert_);
  }

  CredentialsRequest& CACredentialsRequest::Copy(void) const {
    CACredentialsRequest* req = new CACredentialsRequest(Context());
    req->AssignRequest(x509_req_, priv_key_);
    if(cert_)req->Assign(cert_);
    return *req;
  }

  AuthN::Status CACredentialsRequest::MakeRequest(const std::string& conf_file) {
    x509_req_ = X509_REQ_new();
    if(!x509_req_) { return Status(-1); }
    if(!make_request(x509_req_, pub_key_, priv_key_, true, conf_file, subject_name_)) { return Status(-1); }
    return Status(0);
  }

  const X509* CACredentialsRequest::GetCertificate(void) const {
    return cert_;
  }

  void CACredentialsRequest::GetCertificate(std::string& str) const {
    if(!cert_) return;
    BIO *out = BIO_new(BIO_s_mem());
    if(!out) return;
    if(!PEM_write_bio_X509(out,cert_)) { BIO_free_all(out); return; };

    for(;;) {
      char s[256];
      int l = BIO_read(out,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(out);
    return;
  }

  void CACredentialsRequest::GetCertificate(std::ostream& o) const {
    std::string s;
    GetCertificate(s);
    o << s;
  }

  AuthN::Status CACredentialsRequest::Assign(X509 *cert) {
    if(cert) {
      if(cert_) X509_free(cert_);
      cert_ = X509_dup(cert);
    }
    else CredentialRequestLogger.msg(INFO,"Certificate is empty");
    
    return Status(0);
  }

  ProxyCredentialsRequest::ProxyCredentialsRequest(const AuthN::Context& ctx) : CredentialsRequest(ctx) { };

  ProxyCredentialsRequest::~ProxyCredentialsRequest(void) { };

  CredentialsRequest& ProxyCredentialsRequest::Copy(void) const {
    ProxyCredentialsRequest* req = new ProxyCredentialsRequest(Context());
    req->AssignRequest(x509_req_, priv_key_);
    return *req;
  }

  bool ProxyCredentialsRequest::GetLimited(void) {
    return false;
  }

  void ProxyCredentialsRequest::SetLimited(bool) {
  }

  void ProxyCredentialsRequest::GetPolicy(Credentials::Extension& policy) {
    std::string oid;
    bool crit = false;
    std::string ext_val;

    STACK_OF(X509_EXTENSION)* exts = NULL;
    X509_EXTENSION* x509_ext;
    exts = X509_REQ_get_extensions(x509_req_);
    int num = sk_X509_EXTENSION_num(exts);
    for(int i = 0; i < num; i++) {
      int nid;
      x509_ext = sk_X509_EXTENSION_value(exts, i);
      nid = OBJ_obj2nid(X509_EXTENSION_get_object(x509_ext));
      if (nid == NID_proxyCertInfo) {
        if(!parse_extension(x509_ext, crit, oid, ext_val)) return;
      }
    }
    policy.critical = crit;
    policy.oid = oid;
    policy.value = ext_val;
    return;
  }

  void ProxyCredentialsRequest::SetPolicy(const Credentials::Extension& policy) {
    X509_EXTENSION* x509_ext = NULL;
    X509_EXTENSION* pci_ext = NULL;
    STACK_OF(X509_EXTENSION)* exts = NULL;

    int i, nid;
    bool ret = false;

    const EVP_MD* digest = NULL;

    exts = X509_REQ_get_extensions(x509_req_);

    ASN1_OBJECT* pl_obj= NULL;
    ASN1_OCTET_STRING* policy_string = NULL;
    PROXY_CERT_INFO_EXTENSION proxy_info;
    PROXY_POLICY proxy_policy;
    memset(&proxy_info,0,sizeof(proxy_info));
    memset(&proxy_policy,0,sizeof(proxy_policy));
    proxy_info.pcPathLengthConstraint=NULL;
    proxy_info.proxyPolicy=&proxy_policy;
    proxy_policy.policyLanguage=NULL;
    proxy_policy.policy=NULL;

    std::string proxypolicy = policy.value;
    if(!proxypolicy.empty()) {
      pl_obj=OBJ_nid2obj(NID_id_ppl_anyLanguage);  // Proxy with policy
      if(!pl_obj) goto err;
      policy_string=ASN1_OCTET_STRING_new();
      if(!policy_string) goto err;
      ASN1_OCTET_STRING_set(policy_string,(const unsigned char*)(proxypolicy.c_str()),proxypolicy.length());
      proxy_policy.policyLanguage=pl_obj;
      proxy_policy.policy=policy_string;
    } else {
      pl_obj=OBJ_nid2obj(NID_id_ppl_inheritAll);  // Unrestricted proxy
      if(!pl_obj) goto err;
      proxy_policy.policyLanguage=pl_obj;
    }
    //if(X509V3_add1_i2d(&exts, NID_proxyCertInfo,&proxy_info, 1, X509V3_ADD_REPLACE) != 1) 
    x509_ext = X509V3_EXT_i2d(NID_proxyCertInfo, true, &proxy_info);
    sk_X509_EXTENSION_push(exts, x509_ext);

    //We have to delete the existing extensions in X509_REQ, 
    //because X509_REQ_add_extensions_nid has not removed the original 
    //extensions with duplicated nid
    for(i = 0; i<sk_X509_ATTRIBUTE_num(x509_req_->req_info->attributes); i++) {
      X509_REQ_delete_attr(x509_req_, i);
    }

    if(X509_REQ_add_extensions_nid(x509_req_, exts, NID_ext_req) != 1)
      CredentialRequestLogger.msg(ERROR,"Failed to add proxy certinfo extension into X509 request");

    digest = EVP_sha1();
    if(!X509_REQ_sign(x509_req_, (EVP_PKEY*)priv_key_, digest)) { 
      CredentialRequestLogger.msg(ERROR,"Failed to re-sign the x509 request after the extension changed");    
      OpenSSLErrors();
    }

err:
    if(policy_string) ASN1_OCTET_STRING_free(policy_string); policy_string=NULL;
    if(pl_obj)ASN1_OBJECT_free(pl_obj); pl_obj=NULL;
    return;
  }

}//AuthN


