#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <openssl/x509.h>

#include "logger.h"
#include "regexpression.h"
#include "namespacespolicy.h"


namespace AuthN {

  Logger& logger = log();

  const std::string issuername = "TO Issuer";
  const std::string permit  = "PERMIT";
  const std::string deny    = "DENY";
  const std::string subjectname = "Subject";
  const std::string selfissuer = "SELF";
  const char policy_suffix[]   = ".namespaces";

  static void get_line(std::istream& in,std::string& s) {
    for(;;) {
      s.resize(0);
      if(in.fail() || in.eof()) return;
      getline(in,s);
      std::string::size_type p;
      p=s.find_first_not_of(" \t");
      if(p != std::string::npos) s=s.substr(p);
      if((!s.empty()) && (s[0] != '#')) break;
    };
    return;
  }

  static void get_one_rule(std::istream& in,std::string& s) {
    //Combine a statment that is across multiple lines into one string
    s.resize(0);
    std::string s1;
    do {
      s1.resize(0);
      get_line(in,s1);
      std::string::size_type pos;
      pos = s1.find("\\");
      if(pos != std::string::npos) {
        s1 = s1.substr(0, pos); 
        s.append(s1).append(" ");
        continue;
      }
      else { if(!s1.empty())s.append(s1); break; }
    } while(1);
    return;
  }

  static int memicmp(const void *s1, const void *s2, size_t n) {
    const unsigned char *p1 = (const unsigned char *) s1;
    const unsigned char *p2 = (const unsigned char *) s2;
    for (; n-- > 0; p1++, p2++) {
      int c;
      if (*p1 != *p2 && (c = (toupper(*p1) - toupper(*p2)))) return c;
    }
    return 0;
  }

  static int find(const char* s, int s_length, const char* to_find, int to_find_length) {
    int n = s_length - to_find_length;
    int pos = -1;
    if(n == 0 && memicmp(s, to_find, to_find_length) == 0) return 0;
    while( n-- > 0 && memicmp(s, to_find, to_find_length) != 0) {
      ++s; ++pos;
    }
    return pos;
  }

  static void get_word(std::string& s,std::string& word) {
    std::string::size_type w_s;
    std::string::size_type w_e;
    word.resize(0);
    w_s=s.find_first_not_of(" \t");
    if(w_s == std::string::npos) { s.resize(0); return; };
    if(s[w_s] == '\'') {
      w_e=s.find('\'',++w_s);
    } else if(s[w_s] == '"') {
      w_e=s.find('"',++w_s);
    } else {
      w_e=s.find_first_of(" \t",w_s);
    };
    if(w_e == std::string::npos) w_e=s.length();
    word=s.substr(w_s,w_e-w_s);
    if((s[w_e] == '\'') || (s[w_e] == '"')) ++w_e;
    w_s=s.find_first_not_of(" \t",w_e);
    if(w_s == std::string::npos) w_s=w_e;
    s=s.substr(w_s);
    return;
  }

  static bool get_issuer(std::string& s, std::string& issuer) {
    issuer.resize(0);
    int f;
    std::string str;
    f = find(s.c_str(), s.length(), issuername.c_str(), issuername.length());
    if(f != 0) { 
      str = s.substr(f + issuername.length() + 1);
      get_word(str, issuer);
      if(issuer.empty()) {
        logger.msg(WARNING,"Missing issuer in namespaces policy");
        return false;
      }
      return true; 
    }
    return false; 
  }

  static bool get_right(std::string& s, std::string& right) {
    right.resize(0);
    int f;
    std::string str;
    f = find(s.c_str(), s.length(), permit.c_str(), permit.length());
    if(f != 0) {
      str = s.substr(f + 1);
      get_word(str, right);
      right.resize(0); right = "permit";
      return true;
    }
    f = find(s.c_str(), s.length(), deny.c_str(), deny.length());
    if(f != 0) {
      str = s.substr(f + 1);
      get_word(str, right); 
      right.resize(0); right = "deny";
      return true;
    }
    if(right.empty()) {
      logger.msg(WARNING,"Missing right in namespaces policy");
    }
    return 0;
  }

  static bool get_subject(std::string& s, std::list<std::string>& patterns) {
    patterns.resize(0);
    int f;
    std::string str;
    f = find(s.c_str(), s.length(), subjectname.c_str(), subjectname.length());
    if(f != 0) {
      str = s.substr(f + subjectname.length() + 1);
      std::string subject;
      get_word(str,subject);
      if(subject.empty()) {
        logger.msg(WARNING,"Missing subjects in namespaces policy");
        return false;
      }
      patterns.push_back(subject);
      return true;
    }
    return false;     
  }

  static bool match_all(const std::string& issuer_subject,const std::string& subject,const std::string& policy_ca_subject,std::list<std::string> policy_patterns) {
    if(issuer_subject == policy_ca_subject) {
      std::list<std::string>::iterator pattern = policy_patterns.begin();
      for(;pattern!=policy_patterns.end();++pattern) {
        std::string::size_type p = 0;
        for(;;) {
          p=pattern->find('*',p);
          if(p == std::string::npos) break;
          pattern->insert(p,"."); p+=2;
        };
        (*pattern)="^"+(*pattern)+"$";
        AuthN::Utils::RegularExpression re(*pattern);
        bool r = re.match(subject);
        if(r) return true;
      };
    };
    return false;
  }

  static void X509_NAME_to_string(std::string& str,const X509_NAME* name) {
    str.resize(0);
    if(name == NULL) return;
    char* s = X509_NAME_oneline((X509_NAME*)name,NULL,0);
    if(s) {
      str=s;
      OPENSSL_free(s);
    };
    return;
  }

  bool NamespacesPolicy::Match(const X509_NAME* subject) {
    if(!file_) return false;
    std::string subject_str;
    std::string s;
    std::string policy_ca_subject;
    std::string policy_right;
    std::list<std::string> policy_patterns;
    bool right = false;
    bool failure = false;
    X509_NAME_to_string(subject_str, subject);
    bool permit = false;
    for(;;) {
      right = false;
      failure = false;
      get_one_rule(*file_, s);
      if(s.empty()) break;
      logger.msg(INFO, "Get rule get from namespaces file: %s", s.c_str());

      policy_ca_subject.resize(0);
      if(!get_issuer(s, policy_ca_subject)) failure = true;
      //If the "SELF" token is contained.
      //According to the specification: 
      //The (quoted) issuerDirectoryName may be replaced by the single 
      //token “SELF”, in which case this policy applies to the issuer 
      //whose hash corresponds to the hash contained in the namespaces file name
      if(memicmp(policy_ca_subject.c_str(), selfissuer.c_str(), policy_ca_subject.length()) == 0) {
        policy_ca_subject.resize(0);
        policy_ca_subject = issuer_;
      }

      policy_right.resize(0);
      if(!get_right(s, policy_right)) failure = true;
      else { if(policy_right.compare("permit") == 0) right = true; else right = false; }

      policy_patterns.resize(0);
      if(!get_subject(s, policy_patterns)) failure = true;

      if((!policy_ca_subject.empty()) && (!failure)) {
        //According to eugridpma's namespaces format specification: 
        //The policy can either permit or deny the issuer the right 
        //to issue subjects with specific names. A denial overrides 
        //any permissive statements in the same file.
        bool r = match_all(issuer_, subject_str, policy_ca_subject, policy_patterns);
        if(r && right) {
          logger.msg(INFO, "The issuer: %s is permitted to sign the subject: %s", issuer_.c_str(), subject_str.c_str());
          permit = true;
        }
        else if(r && !right) { //Denial overrides
          logger.msg(INFO, "The issuer: %s is explicitly denied to sign the subject: %s", issuer_.c_str(), subject_str.c_str());
          return false;
        }
        else {
          logger.msg(INFO, "The subject: %s and issuer: %s in the verified certificate does not match any namespaces policies files", subject_str.c_str(), issuer_.c_str());
        }
      };
    };
    return permit;
  }

  NamespacesPolicy::NamespacesPolicy(const X509_NAME* issuer_subject,const std::string& ca_path,bool must): file_(NULL) {
    X509_NAME_to_string(issuer_,issuer_subject);
    unsigned long hash = X509_NAME_hash((X509_NAME*)issuer_subject);
    char hash_str[32];
    snprintf(hash_str,sizeof(hash_str)-1,"%08lx",hash);
    hash_str[sizeof(hash_str)-1]=0;
    std::string fname = ca_path+"/"+hash_str+policy_suffix;
    struct stat st;
    if(stat(fname.c_str(),&st) != 0) {
      logger.msg(must?INFO:ERROR, "The namespaces location %s for %s was not found", fname.c_str(), issuer_.c_str());
      return;
    }
    if(!S_ISREG(st.st_mode)) {
      logger.msg(must?INFO:ERROR, "The namespaces location %s for %s is not a regular file", fname.c_str(), issuer_.c_str());
      return;
    }
    file_ = new std::ifstream(fname.c_str());
    if(!(*file_)) { delete file_; file_ = NULL; };
  }

  NamespacesPolicy::~NamespacesPolicy() {
    if(file_) delete file_;
  }

}

