#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>

#include <unistd.h>
#include <openssl/ocsp.h>
#include <openssl/ossl_typ.h>
#include <openssl/rand.h>

#include "opensslutil.h"
#include "logger.h"
#include "dcisec.h"

#define OpenSSLError() \
{ \
        unsigned long errors; \
        char errors_buf[256] = { 0 }; \
        do{ \
                errors = ERR_get_error(); \
                if (errors != 0) { \
                        OCSPClientLogger.msg(ERROR, "OpenSSL error string: %s", ERR_error_string(errors, errors_buf)); \
                } \
        }  while(errors != 0); \
}

namespace AuthN {
namespace OpenSSL {

  Logger& OCSPClientLogger = log();

  static bool make_ocsp_cert(OCSP_REQUEST** req, X509* cert, X509* issuer, 
    X509* signercert, EVP_PKEY* signerkey, STACK_OF(OCSP_CERTID)* ids) {
    OCSP_CERTID* id;
    const EVP_MD* cert_id_md;
    bool ret = false;

    cert_id_md = EVP_sha1();
    if(!issuer) {
      OCSPClientLogger.msg(ERROR, "No issuer certificate specified");
      return false;
    }
    if(!*req) *req = OCSP_REQUEST_new();
    if(!*req) goto end;
    id = OCSP_cert_to_id(cert_id_md, cert, issuer);
    if(!id || !sk_OCSP_CERTID_push(ids, id)) goto end;
    if(!OCSP_request_add0_id(*req, id)) goto end;
    if(OCSP_request_add1_nonce(*req, NULL, -1) != 1) goto end;

    //Sign the request
    if(!OCSP_request_sign(*req, signercert, signerkey, cert_id_md, NULL, 0)) {
      OCSPClientLogger.msg(ERROR, "Failed to sign OCSP request");
      goto end;
    }
    ret = true;

    end:
    if(ret == false) OCSPClientLogger.msg(ERROR, "Failed to create OCSP request");
    return ret;
  }

/*
  static OCSP_RESPONSE* query_responder(BIO* cbio, char* path, STACK_OF(CONF_VALUE)* headers,
         OCSP_REQUEST* req, int req_timeout) {
    int fd;
    int rv;
    int i;
    OCSP_REQ_CTX* ctx = NULL;
    OCSP_RESPONSE* resp = NULL;
    fd_set confds;
    struct timeval tv;

    if(req_timeout != -1) BIO_set_nbio(cbio, 1);
    rv = BIO_do_connect(cbio);
    if((rv <= 0) && ((req_timeout == -1) || !BIO_should_retry(cbio))) {
      OCSPClientLogger.msg(ERROR, "Failed to connect BIO"); return NULL;
    }
    if(BIO_get_fd(cbio, &fd) <= 0) {
      OCSPClientLogger.msg(ERROR, "Failed to get connection fd"); goto err;
    }
    if(req_timeout != -1 && rv <= 0) {
      FD_ZERO(&confds);
      FD_SET(fd, &confds);
      tv.tv_usec = 0;
      tv.tv_sec = req_timeout;
      rv = select(fd + 1, NULL, (fd_set *)&confds, NULL, &tv);
      if(rv == 0) {
        OCSPClientLogger.msg(ERROR, "Connection time out"); return NULL;
      }
    }

    ctx = OCSP_sendreq_new(cbio, path, NULL, -1);
    if(!ctx) return NULL;

    for(i = 0; i < sk_CONF_VALUE_num(headers); i++) {
      CONF_VALUE* hdr = sk_CONF_VALUE_value(headers, i);
      if(!OCSP_REQ_CTX_add1_header(ctx, hdr->name, hdr->value)) goto err;
    }
    if(!OCSP_REQ_CTX_set1_req(ctx, req)) goto err;

    for(;;) {
      rv = OCSP_sendreq_nbio(&resp, ctx);
      if(rv != -1) break;
      if(req_timeout == -1) continue;
      FD_ZERO(&confds);
      FD_SET(fd, &confds);
      tv.tv_usec = 0;
      tv.tv_sec = req_timeout;
      if(BIO_should_read(cbio)) rv = select(fd + 1, (fd_set *)&confds, NULL, NULL, &tv);
      else if(BIO_should_write(cbio)) rv = select(fd + 1, NULL, (fd_set *)&confds, NULL, &tv);
      else {
        OCSPClientLogger.msg(ERROR, "Unexpected retry condition"); goto err;
      }
      if(rv == 0){
        OCSPClientLogger.msg(ERROR, "Connection time out"); break;
      }
      if(rv == -1) {
        OCSPClientLogger.msg(ERROR, "Socket select error"); break;
      }
    }
    err:
    if(ctx) OCSP_REQ_CTX_free(ctx);
    return resp;
  }
*/

  static OCSP_RESPONSE* contact_ocsp_responder(OCSP_REQUEST* req, const std::string& url) {
                        //,STACK_OF(CONF_VALUE) *headers, int req_timeout) {
    BIO* cbio = NULL;
    SSL_CTX* ctx = NULL;
    OCSP_RESPONSE* resp = NULL;
    char *host = NULL, *path = NULL, *port = NULL;
    int use_ssl;
    if(!OCSP_parse_url((char*)(url.c_str()), &host, &port, &path, &use_ssl)) {
      OCSPClientLogger.msg(ERROR, "Failed to parse url of OCSP responder");  goto err;
    }
    if(!path) {
      path = (char*)OPENSSL_malloc(2);
      if(!path) {
        OCSPClientLogger.msg(ERROR, "Failed to allocate memory");  goto err;
      }
      path[0] = '/'; path[1] = '\0';
    }
    cbio = BIO_new_connect(host);
    if(!cbio) {
      OCSPClientLogger.msg(ERROR, "Failed to create connect BIO for sending OCSP request");
      goto err;
    }
    if(port) BIO_set_conn_port(cbio, port);
    if(use_ssl) {
      BIO* sbio;
      ctx = SSL_CTX_new(SSLv23_client_method());
      if(ctx == NULL) {
        OCSPClientLogger.msg(ERROR, "Failed to create SSL context"); goto err;
      }
      SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
      sbio = BIO_new_ssl(ctx, 1);
      cbio = BIO_push(sbio, cbio);
    }
    //resp = query_responder(cbio, path, headers, req, req_timeout);
    if(BIO_do_connect(cbio) <=0) {
      OCSPClientLogger.msg(ERROR, "Failed when connecting BIO"); goto err;
    }
    resp = OCSP_sendreq_bio(cbio, path, req);
    if(!resp) OCSPClientLogger.msg(ERROR, "Failed to query OCSP responder");

    err:
    if(cbio) BIO_free_all(cbio);
    if(ctx) SSL_CTX_free(ctx);
    if(host) OPENSSL_free(host);
    if(port) OPENSSL_free(port);
    if(path) OPENSSL_free(path);
    if(ctx) SSL_CTX_free(ctx);

    return resp;
  }

#define my_CHECKED_I2D_OF(type, i2d) \
    ((i2d_of_void*) (1 ? i2d : ((I2D_OF(type))0)))

#define my_CHECKED_PTR_OF(type, p) \
    ((void*) (1 ? p : (type*)0))

#define my_ASN1_i2d_bio_of(type,i2d,out,x) \
    (ASN1_i2d_bio(my_CHECKED_I2D_OF(type, (int (*)(OCSP_RESPONSE*, unsigned char**))i2d), \
                  out, \
                  (unsigned char*)my_CHECKED_PTR_OF(type, x)))

#define my_i2d_OCSP_RESPONSE_bio(bp,o) my_ASN1_i2d_bio_of(OCSP_RESPONSE,i2d_OCSP_RESPONSE,bp,o)

  static int print_ocsp_summary(BIO *out, OCSP_BASICRESP *bs, OCSP_REQUEST *req,
        std::vector<std::string> names, STACK_OF(OCSP_CERTID) *ids, long nsec, long maxage) {
    OCSP_CERTID* id;
    int i;
    int status, reason;

    ASN1_GENERALIZEDTIME *rev, *thisupd, *nextupd;

    if(!bs || !req || !ids) return 1;
    if(!sk_OCSP_CERTID_num(ids) || (sk_OCSP_CERTID_num(ids) != (int)names.size())) return 1;

    for(i = 0; i < sk_OCSP_CERTID_num(ids); i++) {
      id = sk_OCSP_CERTID_value(ids, i);
      BIO_printf(out, "%s: ", names[i].c_str());

      if(!OCSP_resp_find_status(bs, id, &status, &reason, &rev, &thisupd, &nextupd)) {
        BIO_puts(out, "ERROR: No Status found.\n");
        continue;
      }

      /* Check validity: if invalid write to output BIO so we
       * know which response this refers to.
       */
      if(!OCSP_check_validity(thisupd, nextupd, nsec, maxage)) {
        BIO_puts(out, "WARNING: Status times invalid.\n");
        ERR_print_errors(out);
      }
      BIO_printf(out, "%s\n", OCSP_cert_status_str(status));

      BIO_puts(out, "\tThis Update: ");
      ASN1_GENERALIZEDTIME_print(out, thisupd);
      BIO_puts(out, "\n");

      if(nextupd) {
        BIO_puts(out, "\tNext Update: ");
        ASN1_GENERALIZEDTIME_print(out, nextupd);
        BIO_puts(out, "\n");
      }

      if(status != V_OCSP_CERTSTATUS_REVOKED) continue;
      else return 0;

      if(reason != -1) BIO_printf(out, "\tReason: %s\n", OCSP_crl_reason_str(reason));

      BIO_puts(out, "\tRevocation Time: ");
      ASN1_GENERALIZEDTIME_print(out, rev);
      BIO_puts(out, "\n");
    }

    return 1;
  }

  bool query_ocsp_responder(const std::string& url,
                            const std::string& cert_file, const std::string& issuer_file,
                            const std::string& signercert_file, const std::string& signerkey_file, 
                            const std::string& cadir, const std::string& cafile,
                            std::string& resp_str) {
    X509* cert_to_check = NULL;
    X509* issuer = NULL;
    X509* signer_cert = NULL;
    EVP_PKEY* signer_key = NULL;
    std::string passphrase; //TODO
    bool ret = false;

    loadCertificateFile(cert_file, cert_to_check, NULL);
    if(!cert_to_check) goto end;

    loadCertificateFile(issuer_file, issuer, NULL);
    if(!issuer) goto end;

    loadCertificateFile(signercert_file, signer_cert, NULL);
    if(!signer_cert) goto end;

    loadKeyFile(signerkey_file, signer_key, passphrase);
    if(!signer_key) goto end;

    ret = query_ocsp_responder(url, cert_to_check, issuer, signer_cert, signer_key, cadir, cafile, cert_file, resp_str);

    end:
    if(cert_to_check) X509_free(cert_to_check);
    if(issuer) X509_free(issuer);
    if(signer_cert) X509_free(signer_cert);
    if(signer_key) EVP_PKEY_free(signer_key);

    return ret;
  }

  bool query_ocsp_responder(const std::string& url,
                            X509* cert_to_check, X509* issuer_cert, X509* signer_cert, 
                            EVP_PKEY* signer_key,
                            const std::string& cadir, const std::string& cafile,
                            const std::string& cert_path,
                            std::string& resp_str) {
    OCSP_REQUEST* req = NULL;
    OCSP_RESPONSE* resp = NULL;
    OCSP_BASICRESP* bs = NULL;
    std::vector<std::string> reqnames;
    STACK_OF(OCSP_CERTID)* ids = NULL;
    long nsec = 300; //maximum validity discrepancy in seconds
    long maxage = -1;
    BIO* resp_bio = NULL;
    BIO* out = NULL;
    X509_STORE* verify_store = NULL;
    unsigned long verify_flags = 0;
    int i;
    bool ret = false;

    ids = sk_OCSP_CERTID_new_null();
    if(!make_ocsp_cert(&req, cert_to_check, issuer_cert, signer_cert, signer_key, ids)) goto end;  
    resp = contact_ocsp_responder(req, url);
    if(!resp) goto end;

    reqnames.push_back(cert_path.c_str());

    resp_bio = BIO_new(BIO_s_mem());
    if(!resp_bio) { OCSPClientLogger.msg(ERROR, "Failed to create memory BIO"); goto end; }
    my_i2d_OCSP_RESPONSE_bio(resp_bio, resp);
    //ASN1_i2d_bio_of(OCSP_RESPONSE,i2d_OCSP_RESPONSE,resp_bio,(unsigned char*)resp);
    //ASN1_i2d_bio(CHECKED_I2D_OF(const OCSP_RESPONSE, (int (*)(const OCSP_RESPONSE*, unsigned char**))i2d_OCSP_RESPONSE), resp_bio, (unsigned char*)CHECKED_PTR_OF(const OCSP_RESPONSE, resp));

    resp_str.clear();
    for(;;) {
      char s[256];
      int l = BIO_read(resp_bio,s,sizeof(s));
      if(l <= 0) break;
      resp_str.append(s,l);
    }
    BIO_free(resp_bio);
            
    i = OCSP_response_status(resp);
    if(i != OCSP_RESPONSE_STATUS_SUCCESSFUL) {
      OCSPClientLogger.msg(ERROR, "OCSP Responder gives error: %s (%d)", OCSP_response_status_str(i), i);
      goto end;
    }

    //verify the OCSP response
    verify_store = setup_verify(cafile, cadir); 
    if(!verify_store) goto end;
    bs = OCSP_response_get1_basic(resp);
    if(!bs) { 
      OCSPClientLogger.msg(ERROR, "Failed to parse OCSP respose"); goto end;
    }
    if(req && ((i = OCSP_check_nonce(req, bs)) <= 0)) {
      if(i == -1) OCSPClientLogger.msg(WARNING, "There is no nonce in response");
      else {
        OCSPClientLogger.msg(ERROR, "Failed to verify nonce"); goto end;
      }
    }
    i = OCSP_basic_verify(bs, NULL, verify_store, verify_flags);
    if(i < 0) i = OCSP_basic_verify(bs, NULL, verify_store, 0);
    if(i <= 0) {
      OCSPClientLogger.msg(ERROR, "Failed to verify the response");
      OpenSSLError();
    }
    else OCSPClientLogger.msg(INFO, "Succeeded to verify the response");

    out = BIO_new_fp(stdout, BIO_NOCLOSE);
    if(!print_ocsp_summary(out, bs, req, reqnames, ids, nsec, maxage)) goto end;

    ret = true;

    end:
    if(ids) sk_OCSP_CERTID_free(ids);
    if(req) OCSP_REQUEST_free(req);
    if(resp) OCSP_RESPONSE_free(resp);
    if(bs) OCSP_BASICRESP_free(bs);
    if(out) BIO_free(out);
    if(verify_store) X509_STORE_free(verify_store);

    return ret;
  }

}
}
