#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <cstring>
#include <stdexcept>

#include <openssl/ssl.h>
#include <openssl/ui.h>

#include "opensslutil.h"
#include "logger.h"
#include "namespacespolicy.h"
#include "datetime.h"
#include "dcisec.h"

namespace AuthN {
namespace OpenSSL {

#define OpenSSLError() \
{ \
        unsigned long errors; \
        char errors_buf[256] = { 0 }; \
        do{ \
                errors = ERR_get_error(); \
                if (errors != 0) { \
                        OpenSSLUtilLogger.msg(ERROR, "OpenSSL error string: %s", ERR_error_string(errors, errors_buf)); \
                } \
        }  while(errors != 0); \
}

  Logger& OpenSSLUtilLogger = log();

  static const char DigitalSignature_id[]    = "KeyUsage.digitalSignature";
  static const char NonRepudiation_id[]      = "KeyUsage.nonRepudiation";
  static const char KeyEncipherment_id[]     = "KeyUsage.keyEncipherment";
  static const char DataEncipherment_id[]    = "KeyUsage.dataEncipherment";
  static const char KeyAgreement_id[]        = "KeyUsage.keyAgreement";
  static const char KeyCertificateSign_id[]  = "KeyUsage.keyCertSign";
  static const char CRLSign_id[]             = "KeyUsage.crlSign";
  static const char EncipherOnly_id[]        = "KeyUsage.encipherOnly";
  static const char DecipherOnly_id[]        = "KeyUsage.decipherOnly";
  static const char ServerAuth_id[]          = "1.3.6.1.5.5.7.3.1";
  static const char ClientAuth_id[]          = "1.3.6.1.5.5.7.3.2";
  static const char CodeSigning_id[]         = "1.3.6.1.5.5.7.3.3";
  static const char EmailProtection_id[]     = "1.3.6.1.5.5.7.3.4";
  static const char IPSecEndSystem_id[]      = "1.3.6.1.5.5.7.3.5";
  static const char IPSecTunnel_id[]         = "1.3.6.1.5.5.7.3.6";
  static const char IPSecUser_id[]           = "1.3.6.1.5.5.7.3.7";
  static const char TimeStamping_id[]        = "1.3.6.1.5.5.7.3.8";
  static const char OCSPSigning_id[]         = "1.3.6.1.5.5.7.3.9";

  class CredentialError : public Status {
    public:
      // Constructor
      /** This is the constructor of the CredentialError class.
       * @param what  An explanation of the error.
       */
      CredentialError(const std::string& what="");
  };
  CredentialError::CredentialError(const std::string& what) : Status(-1,what) { }

  AuthN::Utils::Time asn1_to_utctime(const ASN1_UTCTIME *s) {
    if(s == NULL) return -1;
    std::string t_str;
    if(s->type == V_ASN1_UTCTIME) {
      t_str.append("20");
      t_str.append((char*)(s->data));
    }
    else {//V_ASN1_GENERALIZEDTIME
      t_str.append((char*)(s->data));
    }
    return AuthN::Utils::Time(t_str);
  }

  static BIO* OpenFileBIO(const std::string& file) {
    struct stat st;
    if(stat(file.c_str(),&st) != 0) return NULL;
    if(!S_ISREG(st.st_mode)) return NULL;
//    if(!Glib::file_test(file,Glib::FILE_TEST_IS_REGULAR)) return NULL;
    return  BIO_new_file(file.c_str(), "r");
  }

  //Parse the BIO for certificate and get the format of it
  Credformat getFormat(BIO* bio, const bool is_file) {
    Credformat format = CRED_UNKNOWN;
    if(bio == NULL) return format;
    if(is_file) {
      char buf[1];
      char firstbyte;
      int position;
      if((position = BIO_tell(bio))<0 || BIO_read(bio, buf, 1)<=0 || BIO_seek(bio, position)<0) {
        OpenSSLError();
        OpenSSLUtilLogger.msg(ERROR,"Can't get the first byte of input to determine its format");
        return format;
      }
      firstbyte = buf[0];
      // DER-encoded structure (including PKCS12) will start with ASCII 048.
      // Otherwise, it's PEM.
      if(firstbyte==48) {
        //DER-encoded, PKCS12 or DER? firstly parse it as PKCS12 ASN.1,
        //if can not parse it, then it is DER format
        PKCS12* pkcs12 = NULL;
        if((pkcs12 = d2i_PKCS12_bio(bio,NULL)) == NULL){ format=CRED_DER; PKCS12_free(pkcs12); }
        else { format = CRED_PKCS; PKCS12_free(pkcs12); }
        if( BIO_seek(bio, position) < 0 ) {
          OpenSSLError();
          OpenSSLUtilLogger.msg(ERROR,"Can't reset the input");
          return format;
        }
      }
      else { format = CRED_PEM; }
    }
    else {
      unsigned char* bio_str;
      int len;
      len = BIO_get_mem_data(bio, (unsigned char *) &bio_str);
      char firstbyte;
      if(len>0) {
        firstbyte = bio_str[0];
        if(firstbyte==48)  {
          //DER-encoded, PKCS12 or DER? firstly parse it as PKCS12 ASN.1,
          //if can not parse it, then it is DER format
          AutoBIO pkcs12bio(BIO_new_mem_buf(bio_str,len));
          PKCS12* pkcs12 = NULL;
          if((pkcs12 = d2i_PKCS12_bio(pkcs12bio,NULL)) == NULL){
            format=CRED_DER;
          } else {
            format = CRED_PKCS; PKCS12_free(pkcs12);
          }
        } else { format = CRED_PEM; }
      }
      else {
        OpenSSLUtilLogger.msg(ERROR,"Can't get the first byte of input BIO to get its format");
        return format;
      }
    }
    return format;
  }

  void loadCertificateChain(BIO* certchainbio, STACK_OF(X509) **certchain) {
    int n;
    if(certchain == NULL) return;
    if(*certchain == NULL) { *certchain = sk_X509_new_null(); n = 0; }
    else { n = sk_X509_num(*certchain); }

    while(!BIO_eof(certchainbio)){
      X509* tmp = NULL;
      if(!(PEM_read_bio_X509(certchainbio, &tmp, NULL, NULL))){
        ERR_clear_error(); break;
      }
      if(!sk_X509_insert(*certchain, tmp, n)) {
        //std::string str(X509_NAME_oneline(X509_get_subject_name(tmp),0,0));
        X509_free(tmp);
        throw CredentialError("Can not insert cert into certificate's issuer chain");
      }
      ++n;
    }
    if(n==0 && *certchain) { sk_X509_pop_free(*certchain, X509_free); *certchain = NULL; }
  }

  void loadCertificate(BIO* certbio, X509* &x509, STACK_OF(X509) **certchain, const bool is_file) {
    //Parse the certificate
    Credformat format = CRED_UNKNOWN;
    PKCS12* pkcs12 = NULL;
    STACK_OF(X509)* pkcs12_certs = NULL;
    bool get_chain = false;
    if(certchain != NULL) get_chain = true;

    if(certbio == NULL) return;
    format = getFormat(certbio, is_file);
    int n = 0;
    if(get_chain && *certchain) {
      sk_X509_pop_free(*certchain, X509_free);
      *certchain = NULL;
    }

    switch(format) {
      case CRED_PEM:
        OpenSSLUtilLogger.msg(DEBUG,"Certificate format is PEM");
        //Get the certificte, By default, certificate is without passphrase
        //Read certificate
        if(!(PEM_read_bio_X509(certbio, &x509, NULL, NULL))) {
          throw CredentialError("Can not read cert information from BIO");
        }
        if(!get_chain) break;
        //Get the issuer chain
        *certchain = sk_X509_new_null();
        n = 0;
        while(!BIO_eof(certbio)){
          X509 * tmp = NULL;
          if(!(PEM_read_bio_X509(certbio, &tmp, NULL, NULL))){
            ERR_clear_error(); break;
          }
          if(!sk_X509_insert(*certchain, tmp, n)) {
            //std::string str(X509_NAME_oneline(X509_get_subject_name(tmp),0,0));
            X509_free(tmp);
            throw CredentialError("Can not insert cert into certificate's issuer chain");
          }
          ++n;
        }
        break;

      case CRED_DER:
        OpenSSLUtilLogger.msg(DEBUG,"Certificate format is DER");
        x509=d2i_X509_bio(certbio,NULL);
        if(!x509){
          throw CredentialError("Unable to read DER credential from BIO");
        }
        if(!get_chain) break;
        //Get the issuer chain
        *certchain = sk_X509_new_null();
        n = 0;
        while(!BIO_eof(certbio)){
          X509 * tmp = NULL;
          if(!(d2i_X509_bio(certbio, &tmp))){
            ERR_clear_error(); break;
          }
          if(!sk_X509_insert(*certchain, tmp, n)) {
            //std::string str(X509_NAME_oneline(X509_get_subject_name(tmp),0,0));
            X509_free(tmp);
            throw CredentialError("Can not insert cert into certificate's issuer chain");
          }
          ++n;
        }
        break;

      case CRED_PKCS:
        OpenSSLUtilLogger.msg(DEBUG,"Certificate format is PKCS");
        pkcs12 = d2i_PKCS12_bio(certbio, NULL);
        if(pkcs12){
          char password[100];
          EVP_read_pw_string(password, 100, "Enter Password for PKCS12 certificate:", 0);
          if(!PKCS12_parse(pkcs12, password, NULL, &x509, &pkcs12_certs)) {
            if(pkcs12) PKCS12_free(pkcs12);
            OpenSSLError();
            throw CredentialError("Can not parse PKCS12 file");
          }
        }
        else{
          throw CredentialError("Can not read PKCS12 credential from BIO");
        }
        if(!get_chain) break;
        if (pkcs12_certs && sk_X509_num(pkcs12_certs)){
          X509* tmp;
          for (n = 0; n < sk_X509_num(pkcs12_certs); n++) {
            tmp = X509_dup(sk_X509_value(pkcs12_certs, n));
            sk_X509_insert(*certchain, tmp, n);
          }
        }
        if(pkcs12) { PKCS12_free(pkcs12); }
        if(pkcs12_certs) { sk_X509_pop_free(pkcs12_certs, X509_free); }

        break;

      default:
        OpenSSLUtilLogger.msg(DEBUG,"Certificate format is unknown");
        break;
     } // end switch

    if(n==0 && *certchain) { sk_X509_pop_free(*certchain, X509_free); *certchain = NULL; }
  }

  void loadCertificateFile(const std::string& certfile, X509* &x509, STACK_OF(X509) **certchain) {
    BIO* b = OpenFileBIO(certfile);
    if(!b) {
        OpenSSLUtilLogger.msg(ERROR,"Can not find certificate file: %s", certfile.c_str());
        throw CredentialError("Can not find certificate file");
    }
    AutoBIO certbio(b);
    if(!certbio){
      OpenSSLUtilLogger.msg(ERROR,"Can not read certificate file: %s", certfile.c_str());
      OpenSSLError();
      throw CredentialError("Can not read certificate file");
    }
    loadCertificate(certbio,x509,certchain, true);
  }

  int passwordcb(char *buf, int bufsiz, int verify, void *cb_tmp) {
    UI *ui = NULL;
    int res = 0;
    const char *prompt_info = NULL;
    const char *password = NULL;
    PW_CB_DATA *cb_data = (PW_CB_DATA *)cb_tmp;

    if (bufsiz < 1) return res;
    if (cb_data) {
      if (cb_data->password)
        password = (const char*)(cb_data->password);
      if (cb_data->prompt_info)
        prompt_info = cb_data->prompt_info;
    }

    if (password) {
      res = strlen(password);
      if (res > (bufsiz-1))
        res = bufsiz-1;
      if(buf) {
        memcpy(buf, password, res);
        buf[res] = 0;
      }
      return res;
    }

    //ui = UI_new_method(ui_method);
    ui = UI_new();
    if (ui) {
      int ok = 0;
      char *buf1 = NULL;
      char *buf2 = NULL;
      int ui_flags = 0;
      char *prompt = NULL;

      prompt = UI_construct_prompt(ui, "pass phrase", cb_data->prompt_info);

      ui_flags |= UI_INPUT_FLAG_DEFAULT_PWD;
//      UI_ctrl(ui, UI_CTRL_PRINT_ERRORS, 1, 0, 0);

      if (ok >= 0) {
        ok = -1;
        if((buf1 = (char *)OPENSSL_malloc(bufsiz)) != NULL) {
          memset(buf1,0,(unsigned int)bufsiz);
          ok = UI_add_input_string(ui,prompt,ui_flags,buf1,PASS_MIN_LENGTH,PASS_MAX_LENGTH);
        }
      }
      if (ok >= 0 && verify) {
        ok = -1;
        if((buf2 = (char *)OPENSSL_malloc(bufsiz)) != NULL) {
          memset(buf2,0,(unsigned int)bufsiz);
          ok = UI_add_verify_string(ui,prompt,ui_flags,buf2,PASS_MIN_LENGTH,PASS_MAX_LENGTH, buf);
        }
      }
      if (ok >= 0)
        do{
          ok = UI_process(ui);
          if(ok == -2) break; // Abort request
          if(ok == -1) { // Password error
            unsigned long errcode = ERR_get_error();
            const char* errstr = ERR_reason_error_string(errcode);
            if(errstr == NULL) {
              OpenSSLUtilLogger.msg(ERROR, "error code %lu", errcode);
            } else if(strstr(errstr,"result too small")) {
              OpenSSLUtilLogger.msg(ERROR, "Password is too short, need at least %u charcters",PASS_MIN_LENGTH);
            } else if(strstr(errstr,"result too large")) {
              OpenSSLUtilLogger.msg(ERROR, "Password is too long, need at most %u characters",PASS_MAX_LENGTH);
            } else {
              OpenSSLUtilLogger.msg(ERROR, std::string(errstr));
            };
          };
        }while (ok < 0 && UI_ctrl(ui, UI_CTRL_IS_REDOABLE, 0, 0, 0));

      if (buf2){
        memset(buf2,0,(unsigned int)bufsiz);
        OPENSSL_free(buf2);
      }

      if (ok >= 0) {
        if(buf1) {
          buf1[bufsiz-1] = 0;
          res = strlen(buf1);
          if(buf) memcpy(buf,buf1,res+1);
        }
      }

      if (buf1){
        memset(buf1,0,(unsigned int)bufsiz);
        OPENSSL_free(buf1);
      }

      if (ok == -1){
        OpenSSLUtilLogger.msg(ERROR, "User interface error");
        OpenSSLError();
        if(buf) memset(buf,0,(unsigned int)bufsiz);
        res = 0;
      }
      if (ok == -2) {
        if(buf) memset(buf,0,(unsigned int)bufsiz);
        res = 0;
      }
      UI_free(ui);
      OPENSSL_free(prompt);
    }

    return res;
  }

  void loadKey(BIO* keybio, EVP_PKEY* &pkey, const std::string& passphrase, const std::string& prompt_info, const bool is_file) {
    //Read key
    Credformat format;
    PKCS12* pkcs12 = NULL;

    if(keybio == NULL) return;
    format = getFormat(keybio, is_file);
    std::string prompt;
    switch(format){
      case CRED_PEM:
        PW_CB_DATA cb_data;
        cb_data.password = (passphrase.empty()) ? NULL : (void*)(passphrase.c_str());
        cb_data.prompt_info = prompt_info.empty() ? NULL : prompt_info.c_str();
        if(!(pkey = PEM_read_bio_PrivateKey(keybio, NULL, passwordcb, &cb_data))) {
          int reason = ERR_GET_REASON(ERR_peek_error());
          if(reason == PEM_R_BAD_BASE64_DECODE) 
            OpenSSLUtilLogger.msg(ERROR, "Can not read PEM private key: probably bad password");
          if(reason == PEM_R_BAD_DECRYPT)
            OpenSSLUtilLogger.msg(ERROR, "Can not read PEM private key: failed to decrypt");
          if(reason == PEM_R_BAD_PASSWORD_READ)
            OpenSSLUtilLogger.msg(ERROR, "Can not read PEM private key: failed to obtain password");
          if(reason == PEM_R_PROBLEMS_GETTING_PASSWORD)
            OpenSSLUtilLogger.msg(ERROR,"Can not read PEM private key: failed to obtain password");
          OpenSSLUtilLogger.msg(ERROR, "Can not read PEM private key");
        }
        break;

      case CRED_DER:
        pkey=d2i_PrivateKey_bio(keybio, NULL);
        break;

      case CRED_PKCS:
        pkcs12 = d2i_PKCS12_bio(keybio, NULL);
        if(pkcs12) {
          char password[100];
          EVP_read_pw_string(password, 100, "Enter Password for PKCS12 certificate:", 0);
          if(!PKCS12_parse(pkcs12, password, &pkey, NULL, NULL)) {
            throw CredentialError("Can not parse PKCS12");
          }
          PKCS12_free(pkcs12);
        }
        break;

      default:
        break;
    }
  }

  void loadKeyFile(const std::string& keyfile, EVP_PKEY* &pkey, const std::string& passphrase) {
    BIO* b = OpenFileBIO(keyfile);
    if(!b) {
        OpenSSLUtilLogger.msg(ERROR,"Can not find key file: %s", keyfile.c_str());
        throw CredentialError("Can not find key file");
    }
    AutoBIO keybio(b);
    if(!keybio){
      OpenSSLUtilLogger.msg(ERROR,"Can not open key file %s", keyfile.c_str());
      OpenSSLError();
      throw CredentialError("Can not open key file " + keyfile);
    }
    loadKey(keybio,pkey,passphrase, keyfile, true);
  }

  void loadCRLFile(const std::string& crlfile, X509_CRL* &crl, bool is_pem) {
    BIO* b = OpenFileBIO(crlfile);
    if(!b) {
      OpenSSLUtilLogger.msg(ERROR,"Can not find crl file: %s", crlfile.c_str());
      return;
    }
    AutoBIO crlbio(b);
    if(!crlbio){
      OpenSSLUtilLogger.msg(ERROR,"Can not open crl file %s", crlfile.c_str());
      OpenSSLError(); return;
    }

    if(!is_pem) crl = d2i_X509_CRL_bio(crlbio, NULL);
    else if (is_pem) crl = PEM_read_bio_X509_CRL(crlbio,NULL,NULL,NULL);
    if(crl == NULL) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to load CRL");
      OpenSSLError();
    }
    return;
  }

  int get_proxy_auth_ex_data_idx(void) {
    static volatile int idx = -1;
    if (idx < 0) {
      CRYPTO_w_lock(CRYPTO_LOCK_X509_STORE);
      if (idx < 0) {
        idx = X509_STORE_CTX_get_ex_new_index(0,
              //(void*)"for verify callback", NULL,NULL,NULL);
              (void*)"AuthN Validator", NULL,NULL,NULL);
      }
      CRYPTO_w_unlock(CRYPTO_LOCK_X509_STORE);
    }
    return idx;
  }

  static KeyUsages get_cert_key_usage(X509_EXTENSION *ex) {
    KeyUsages keyusages;
    int bit_table[9] = {
            DigitalSignature,
            NonRepudiation,
            KeyEncipherment,
            DataEncipherment,
            KeyAgreement,
            KeyCertificateSign,
            CRLSign,
            EncipherOnly,
            DecipherOnly
    };

    ASN1_BIT_STRING *keyusage = (ASN1_BIT_STRING *)X509V3_EXT_d2i(ex);
    for(int n = 0; n < 9; ++n) {
      if(ASN1_BIT_STRING_get_bit(keyusage, n)) 
      {
        keyusages += KeyUsage((KeyUsageType)bit_table[n]);
      }
    }
    ASN1_BIT_STRING_free(keyusage);
    return keyusages;
  }

  static KeyUsages get_cert_ext_key_usage(X509_EXTENSION *ex) {
    KeyUsages keyusages;
    EXTENDED_KEY_USAGE *extkeyusage = (EXTENDED_KEY_USAGE *)X509V3_EXT_d2i(ex);
    for(int n = 0; n < sk_ASN1_OBJECT_num(extkeyusage); ++n) {
      ASN1_OBJECT *obj = sk_ASN1_OBJECT_value(extkeyusage, n);
      int nid = OBJ_obj2nid(obj);
      if(nid == NID_undef) continue;

      int t = -1;
      switch(nid) {
        case NID_server_auth: t = ServerAuth; break;
        case NID_client_auth: t = ClientAuth; break;
        case NID_code_sign: t = CodeSigning; break;
        case NID_email_protect: t = EmailProtection; break;
        case NID_ipsecEndSystem: t = IPSecEndSystem; break;
        case NID_ipsecTunnel: t = IPSecTunnel; break;
        case NID_ipsecUser: t = IPSecUser; break;
        case NID_time_stamp: t = TimeStamping; break;
        case NID_OCSP_sign: t = OCSPSigning; break;
      };
      if(t == -1) continue;
      keyusages += KeyUsage((KeyUsageType)t);
    }
    sk_ASN1_OBJECT_pop_free(extkeyusage, ASN1_OBJECT_free);
    return keyusages;
  }

  enum ConstraintBit {
    Bit_DigitalSignature   = 0,
    Bit_NonRepudiation     = 1,
    Bit_KeyEncipherment    = 2,
    Bit_DataEncipherment   = 3,
    Bit_KeyAgreement       = 4,
    Bit_KeyCertificateSign = 5,
    Bit_CRLSign            = 6,
    Bit_EncipherOnly       = 7,
    Bit_DecipherOnly       = 8
  };

  static X509_EXTENSION* new_cert_key_usage(KeyUsages& keyusages) {
    ASN1_BIT_STRING *keyusage = 0;
    for(int n = 0; n < keyusages.Size(); ++n) {
      int bit = -1;
      switch(keyusages[n].type()) {
        case DigitalSignature: bit = Bit_DigitalSignature; break;
        case NonRepudiation: bit = Bit_NonRepudiation; break;
        case KeyEncipherment: bit = Bit_KeyEncipherment; break;
        case DataEncipherment: bit = Bit_DataEncipherment; break;
        case KeyAgreement: bit = Bit_KeyAgreement; break;
        case KeyCertificateSign: bit = Bit_KeyCertificateSign; break;
        case CRLSign: bit = Bit_CRLSign; break;
        case EncipherOnly: bit = Bit_EncipherOnly; break;
        case DecipherOnly: bit = Bit_DecipherOnly; break;
        default: break;
      }
      if(bit != -1) {
        if(!keyusage)
          keyusage = ASN1_BIT_STRING_new();
        ASN1_BIT_STRING_set_bit(keyusage, bit, 1);
      }
    }
    if(!keyusage) return NULL;

    X509_EXTENSION *ex = X509V3_EXT_i2d(NID_key_usage, 1, keyusage);
    ASN1_BIT_STRING_free(keyusage);
    return ex;
  }

  static X509_EXTENSION* new_cert_ext_key_usage(KeyUsages& keyusages) {
    EXTENDED_KEY_USAGE *extkeyusage = NULL;
    for(int n = 0; n < keyusages.Size(); ++n) {
      int nid = -1;
      switch(keyusages[n].type()) {
        case ServerAuth: nid = NID_server_auth; break;
        case ClientAuth: nid = NID_client_auth; break;
        case CodeSigning: nid = NID_code_sign; break;
        case EmailProtection: nid = NID_email_protect; break;
        case IPSecEndSystem: nid = NID_ipsecEndSystem; break;
        case IPSecTunnel: nid = NID_ipsecTunnel; break;
        case IPSecUser: nid = NID_ipsecUser; break;
        case TimeStamping: nid = NID_time_stamp; break;
        case OCSPSigning: nid = NID_OCSP_sign; break;
        default: break;
      }
      if(nid != -1) {
        if(!extkeyusage)
          extkeyusage = sk_ASN1_OBJECT_new_null();
        ASN1_OBJECT *obj = OBJ_nid2obj(nid);
        sk_ASN1_OBJECT_push(extkeyusage, obj);
      }
    }
    if(!extkeyusage) return NULL;

    X509_EXTENSION *ex = X509V3_EXT_i2d(NID_ext_key_usage, 0, extkeyusage);
    sk_ASN1_OBJECT_pop_free(extkeyusage, ASN1_OBJECT_free);
    return ex;
  }

  static KeyUsages get_cert_key_usage(X509* cert) {
    KeyUsages keyusages;
    int index;
    index = X509_get_ext_by_NID(cert, NID_key_usage, -1);
    if(index != -1) {
      X509_EXTENSION *ex = X509_get_ext(cert, index);
      if(ex) {
        keyusages = get_cert_key_usage(ex);
      }
    }
    return keyusages;
  }

  static KeyUsages get_cert_ext_key_usage(X509* cert) {
    KeyUsages keyusages;
    int index;
    index = X509_get_ext_by_NID(cert, NID_ext_key_usage, -1);
    if(index != -1) {
      X509_EXTENSION *ex = X509_get_ext(cert, index);
      if(ex) {
        keyusages = get_cert_ext_key_usage(ex);
      }
    }
    return keyusages;
  }

/*
  static KeyUsages get_key_usage(X509* cert) {
    KeyUsages keyusages;
    int index;
    index = X509_get_ext_by_NID(cert, NID_key_usage, -1);
    if(index != -1) {
      X509_EXTENSION *ex = X509_get_ext(cert, index);
      if(ex) {
        keyusages = get_cert_key_usage(ex);
      }
    }
    index = X509_get_ext_by_NID(cert, NID_ext_key_usage, -1);
    if(index != -1) {
      X509_EXTENSION *ex = X509_get_ext(cert, index);
      if(ex) {
        keyusages += get_cert_ext_key_usage(ex);
      }
    }

    return keyusages;
  }
*/

   /* from section 4.2 of RFC 3820:
   Application making authorization decisions based on the contents of
   the proxy certificate key usage or extended key usage extensions MUST
   examine the list of key usage, extended key usage and proxy policies
   resulting from proxy certificate path validation and determine the
   effective key usage functions of the proxy certificate as follows:

   *  If a certificate is a proxy certificate with a proxy policy of
      id-ppl-independent or an end entity certificate, the effective key
      usage functions of that certificate is as defined by the key usage
      and extended key usage extensions in that certificate.  The key
      usage functionality of the issuer has no bearing on the effective
      key usage functionality.

   *  If a certificate is a proxy certificate with a policy other than
      id-ppl-independent, the effective key usage and extended key usage
      functionality of the proxy certificate is the intersection of the
      functionality of those extensions in the proxy certificate and the
      effective key usage functionality of the proxy issuer.
  */
  int check_issued(X509_STORE_CTX* ctx, X509* x, X509* issuer) {
    int ret;
    ret = X509_check_issued(issuer, x);
    if (ret == X509_V_OK) {
      if(X509_cmp(x,issuer) != 0) {
        /*If the the 'issuer' is the issuer of 'x', not self-signed.
          And if the NID_proxyCertInfo of 'x' is not id-ppl-independent,
          (i.e. 'x' is a proxy certificate with a policy other 
          than id-ppl-independent). According to RFC3820, "the effective 
          key usage and extended key usage functionality of the proxy 
          certificate is the intersection of the functionality of 
          those extensions in the proxy certificate and the
          effective key usage functionality of the proxy issuer"
         */
        PROXY_CERT_INFO_EXTENSION *pci = (PROXY_CERT_INFO_EXTENSION*)X509_get_ext_d2i(x, NID_proxyCertInfo,
                                         NULL, NULL);
        if(pci != NULL) {
          int i = OBJ_obj2nid(pci->proxyPolicy->policyLanguage);
          if(i != NID_Independent) {
            KeyUsages ku1 = get_cert_key_usage(x);
            KeyUsages ku2 = get_cert_key_usage(issuer);
            ku1 = ku1.Intersection(ku2);
            KeyUsages ku3 = get_cert_ext_key_usage(x);
            KeyUsages ku4 = get_cert_ext_key_usage(issuer);
            ku3 = ku3.Intersection(ku4);
            X509_EXTENSION* ex;
            ex = new_cert_key_usage(ku1);
            if(ex) {
              X509_add_ext(x, ex, -1);
              X509_EXTENSION_free(ex);
            }
            ex = new_cert_ext_key_usage(ku3);
            if(ex) {
              X509_add_ext(x, ex, -1);
              X509_EXTENSION_free(ex);
            }
          }
        }
      }
      return 1;
    }

    /* If we haven't asked for issuer errors don't set ctx */
    if (!(ctx->param->flags & X509_V_FLAG_CB_ISSUER_CHECK))
      return 0;

    ctx->error = ret;
    ctx->current_cert = x;
    ctx->current_issuer = issuer;
    return ctx->verify_cb(0, ctx);
    return 0;
  }

  int verify_callback(int preverify_ok, X509_STORE_CTX *ctx) {
    X509* current_cert = X509_STORE_CTX_get_current_cert(ctx);
    int err = X509_STORE_CTX_get_error(ctx);
    int err_depth = X509_STORE_CTX_get_error_depth(ctx);
    char buf[256];
    X509_NAME_oneline(X509_get_subject_name(current_cert),buf, sizeof buf);
    if (!preverify_ok)
      OpenSSLUtilLogger.msg(ERROR,"Verification error number: %d; error string: %s; error depth: %d; subject: %s", err,X509_verify_cert_error_string(err),err_depth,buf);

    OpenSSLUtilLogger.msg(INFO, "Subject name of current X509 to be verified: %s", buf);
    std::string sub_name = buf; 

    if (preverify_ok == 0) {
      switch (err) {
        case X509_V_ERR_UNABLE_TO_GET_CRL: {
          // Missing CRL is not an error (TODO: make it configurable)
          // Consider that to be a policy of site like Globus does
#ifdef HAVE_OPENSSL_X509_VERIFY_PARAM
          if(ctx->param) {
            X509_VERIFY_PARAM_clear_flags(ctx->param,X509_V_FLAG_CRL_CHECK);
            preverify_ok=1;
            if(preverify_ok == 1) X509_STORE_CTX_set_error(ctx,X509_V_OK);
          }
#else
#if OPENSSL_VERSION_NUMBER < 0x00908000
          ctx->flags &= ~X509_V_FLAG_CRL_CHECK;
#else
          ctx->param->flags &= ~X509_V_FLAG_CRL_CHECK;
#endif
          //set_flag_STORE_CTX(ctx,get_flag_STORE_CTX(ctx) | FLAG_CRL_DISABLED);
          //ok=1;
          //if(ok == 1) X509_STORE_CTX_set_error(ctx,X509_V_OK);
#endif
        }; break;
        default: {
          OpenSSLUtilLogger.msg(ERROR,"Verification error number: %d; error string: %s", err, X509_verify_cert_error_string(err));
        }; break;
      }
    }

    /**The proxy related checking goes as the following. 
       Note that the openssl is assumpted to be able to recognize proxy cert info extension;
       We are trying to strictly follow RFC 3820 as much as possible.
    */
     ValidationContext* vctx  = (ValidationContext*)X509_STORE_CTX_get_ex_data(ctx,
                                  get_proxy_auth_ex_data_idx());
    if (preverify_ok == 1) {
      if (current_cert->ex_flags & EXFLAG_PROXY) {
        ProxyPolicyList* proxy_policy_list = &(vctx->ppl);
        if(vctx != NULL) {
          PROXY_CERT_INFO_EXTENSION *pci = (PROXY_CERT_INFO_EXTENSION*)X509_get_ext_d2i(current_cert, NID_proxyCertInfo,
                                           NULL, NULL);
          switch (OBJ_obj2nid(pci->proxyPolicy->policyLanguage)) {
          /**RFC 3820:
          The policyLanguage field has two values of special importance,
          defined in Appendix A, that MUST be understood by all parties
          accepting Proxy Certificates:

         *id-ppl-inheritAll indicates that this is an unrestricted proxy
          that inherits all rights from the issuing PI.  An unrestricted
          proxy is a statement that the Proxy Issuer wishes to delegate all
          of its authority to the bearer (i.e., to anyone who has that proxy
          certificate and can prove possession of the associated private
          key).  For purposes of authorization, this an unrestricted proxy
          effectively impersonates the issuing PI.

         *id-ppl-independent indicates that this is an independent proxy
          that inherits no rights from the issuing PI.  This PC MUST be
          treated as an independent identity by relying parties.  The only
          rights this PC has are those granted explicitly to it.

          For either of the policyLanguage values listed above, the policy
          field MUST NOT be present.
          */
            case NID_Independent:
            /**If there is no explicit rights granted (usually by pulling the 
              rights from some database) to this particular proxy
              certificate, the policy (rights) is set to be empty.
              this particular proxy certificate.
              which means the subsequent proxy is void of any rights, even 
              if it is with pci nid id-ppl-inheritAll
            */
              (*proxy_policy_list).clear();
              break;

            case NID_id_ppl_inheritAll:
            /* This is basically a NOP, we simply let the current
             rights stand as they are. */
              break;

            default:
              ProxyPolicy pp;
              pp.subject_name = sub_name; 
              if((pci->proxyPolicy) && (pci->proxyPolicy->policy) &&
                 (pci->proxyPolicy->policy->data))
                pp.policy = std::string((const char*)(pci->proxyPolicy->policy->data), pci->proxyPolicy->policy->length);
              (*proxy_policy_list).push_back(pp);
              break;
          }	
          PROXY_CERT_INFO_EXTENSION_free(pci);
        }
      }
    } 

    if(preverify_ok == 1) {
      //check CA policy matching
      char* subject_name = X509_NAME_oneline(X509_get_subject_name(current_cert),NULL,0);
      // eugridpma namespaces policy
      // Do not apply to proxies and self-signed CAs.
#ifdef HAVE_OPENSSL_PROXY
      int pos = X509_get_ext_by_NID(current_cert, NID_proxyCertInfo, -1);
      if(pos < 0)
#endif
      {
        //Search the namespace policy file
        NamespacesPolicy* in = NULL;
        if(vctx->validation_mode & Validator::ValidationNamespaceCheck) {
          in = new NamespacesPolicy(X509_get_issuer_name(current_cert), vctx->ca_path, true);
          if(!*in) OpenSSLUtilLogger.msg(ERROR, "Failed to open namespaces policy");
        }
        else if(vctx->validation_mode & Validator::ValidationNamespaceLax) {
          in = new NamespacesPolicy(X509_get_issuer_name(current_cert), vctx->ca_path, false);
          if(!*in) { delete in; in = NULL; }; // No suitable namespace => no check
        }
        if(in) {
          if(!in->Match(X509_get_subject_name(current_cert))) {
            OpenSSLUtilLogger.msg(ERROR,"Certificate %s failed namespaces policy", subject_name);
            preverify_ok=0;
            X509_STORE_CTX_set_error(ctx,X509_V_ERR_SUBJECT_ISSUER_MISMATCH);
          };
        };
        delete in;
      };

      //Check the left validity time of the peer certificate;
      //Give warning if the certificate is going to be expired
      //in a while of time
      AuthN::Utils::Time exptime = asn1_to_utctime(X509_get_notAfter(current_cert));
      if(exptime <= AuthN::Utils::Time()) {
        OpenSSLUtilLogger.msg(WARNING,"Certificate %s already expired", subject_name);
      } else {
        AuthN::Utils::Period timeleft = exptime - AuthN::Utils::Time();
#ifdef HAVE_OPENSSL_PROXY
        int pos = X509_get_ext_by_NID(current_cert,NID_proxyCertInfo,-1);
#else
        int pos = -1;
#endif
        //for proxy certificate, give warning 1 hour in advance
        //for EEC certificate, give warning 5 days in advance
        if(((pos < 0) && (timeleft <= 5*24*3600)) ||
           (timeleft <= 3600)) {
          OpenSSLUtilLogger.msg(WARNING,"Certificate %s will expire in %s", subject_name, ((std::string)timeleft).c_str());
        }
      }
      OPENSSL_free(subject_name);
    };
 
    return(preverify_ok);
  }

  static bool KeyUsageTypeExtended(KeyUsageType k) {
    switch(k) {
      case DigitalSignature:
      case NonRepudiation:
      case KeyEncipherment:
      case DataEncipherment:
      case KeyAgreement:
      case KeyCertificateSign:
      case CRLSign:
      case EncipherOnly:
      case DecipherOnly:
        return false;
      default:
        break;
    }
    return true;
  }

  static int KeyUsageIdToType(const std::string &id) {
    if(id == DigitalSignature_id) 
      return DigitalSignature;
    else if(id == NonRepudiation_id) 
      return NonRepudiation;
    else if(id == KeyEncipherment_id)
      return KeyEncipherment;
    else if(id == DataEncipherment_id)
      return DataEncipherment;
    else if(id == KeyAgreement_id)
      return KeyAgreement;
    else if(id == KeyCertificateSign_id)
      return KeyCertificateSign;
    else if(id == CRLSign_id)
      return CRLSign;
    else if(id == EncipherOnly_id)
      return EncipherOnly;
    else if(id == DecipherOnly_id)
      return DecipherOnly;
    else if(id == ServerAuth_id)
      return ServerAuth;
    else if(id == ClientAuth_id)
      return ClientAuth;
    else if(id == CodeSigning_id)
      return CodeSigning;
    else if(id == EmailProtection_id)
      return EmailProtection;
    else if(id == IPSecEndSystem_id)
      return IPSecEndSystem;
    else if(id == IPSecTunnel_id)
      return IPSecTunnel;
    else if(id == IPSecUser_id)
      return IPSecUser;
    else if(id == TimeStamping_id)
      return TimeStamping;
    else if(id == OCSPSigning_id)
      return OCSPSigning;
    else
      return -1;
  }

  static std::string KeyUsageTypeToId(KeyUsageType k) {
    std::string out;
    switch(k) {
      case DigitalSignature:   out = DigitalSignature_id; break;
      case NonRepudiation:     out = NonRepudiation_id; break;
      case KeyEncipherment:    out = KeyEncipherment_id; break;
      case DataEncipherment:   out = DataEncipherment_id; break;
      case KeyAgreement:       out = KeyAgreement_id; break;
      case KeyCertificateSign: out = KeyCertificateSign_id; break;
      case CRLSign:            out = CRLSign_id; break;
      case EncipherOnly:       out = EncipherOnly_id; break;
      case DecipherOnly:       out = DecipherOnly_id; break;
      case ServerAuth:         out = ServerAuth_id; break;
      case ClientAuth:         out = ClientAuth_id; break;
      case CodeSigning:        out = CodeSigning_id; break;
      case EmailProtection:    out = EmailProtection_id; break;
      case IPSecEndSystem:     out = IPSecEndSystem_id; break;
      case IPSecTunnel:        out = IPSecTunnel_id; break;
      case IPSecUser:          out = IPSecUser_id; break;
      case TimeStamping:       out = TimeStamping_id; break;
      case OCSPSigning:        out = OCSPSigning_id; break;
    }
    return out;
  }
 
  KeyUsage::KeyUsage(KeyUsageType type) {
    extended_ = KeyUsageTypeExtended(type);
    type_ = type;
    id_ = KeyUsageTypeToId(type);
  }

  KeyUsage::KeyUsage(const std::string& id, bool is_extended) {
    extended_ = is_extended;
    type_ = KeyUsageIdToType(id);
    id_ = id;
  }

  bool KeyUsage::operator<(const KeyUsage &other) const {
    // sort by knowns (in enum order), then by ids (in string order)
    if(type_ != -1) {
      if(other.type_ == -1) return true;
      else if(type_ < other.type_) return true;
      else return false;
    }
    else {
      if(other.type_ != -1) return false;
      else if(id_ < other.id_) return true;
      else return false;
    }
  }

  bool KeyUsage::operator==(const KeyUsage &other) const {
    if(type_!= -1 && other.type_ != -1) {
      if(type_ != other.type_) return false;
    }
    else {
      if(id_ != other.id_) return false;
    }
    if(extended_ != other.extended_) return false;
    return true;
  }


  KeyUsages& KeyUsages::operator+= ( const KeyUsages & other ) {
    //for(std::list<KeyUsage>::const_iterator i=other.keyusages.begin(); i!=other.keyusages.end(); i++) { 
    for(int i = 0; i<other.Size(); i++) {
      int j = 0;
      for(; j<(*this).Size(); j++) {
        if((*this)[j] == other[i]) break;
      }       
      if(j == (*this).Size()) keyusages.push_back(other[i]);
    }
    return *this;
  }

  KeyUsages& KeyUsages::operator+= ( const KeyUsage& value ) { 
    //keyusages.push_back(value); return *this; 
    int j = 0;
    for(; j<(*this).Size(); j++) {
      if((*this)[j] == value) break;
    } 
    if(j == (*this).Size()) keyusages.push_back(value);
    return *this;
  }

  KeyUsages& KeyUsages::Intersection( const KeyUsages & other ) {
    std::list<KeyUsage> tmp_kus;
    for(int i = 0; i<other.Size(); i++) {
      int j = 0;
      for(; j<(*this).Size(); j++) {
        if((*this)[j] == other[i]) break;
      }
      if(j != (*this).Size()) tmp_kus.push_back(other[i]);
    }
    keyusages.clear();
    keyusages = tmp_kus;
    return *this;
  }
  
  KeyUsages::operator std::string(void) const {
    std::string ret;
    for(int i = 0; i<(*this).Size(); i++) {
      if(i==0) ret.append((*this)[i].id());
      else ret.append(", ").append((*this)[i].id());
    }
    return ret;
  }

  std::string GetOpenSSLError(int err) {
    std::string errstr;
    unsigned long errors;
    char errors_buf[256] = { 0 };
    if(err == 0) {
      errors = ERR_get_error();
    } else {
      errors = err;
    };
    while(errors != 0){
      ERR_error_string_n(errors,errors_buf,sizeof(errors_buf));
      errors_buf[sizeof(errors_buf)-1] = 0;
      errstr += errors_buf;
      errstr += "\n";
      errors = ERR_get_error();
    };
    return errstr;
  }

  bool parse_extension(X509_EXTENSION* x509_ext, bool& crit, std::string& oid, std::string& ext_val) {
    char buf[256];
    int buflen;
    buflen = OBJ_obj2txt(buf, sizeof(buf), X509_EXTENSION_get_object(x509_ext), 0);
    if(buflen < 0) { OpenSSLUtilLogger.msg(ERROR,"Failed to call OBJ_obj2txt to convert obj to txt"); return false; }
    else oid = buf;

    if(X509_EXTENSION_get_critical(x509_ext)) crit=true;

    BIO* bio = NULL;
    bio = BIO_new(BIO_s_mem());
    //ASN1_OBJECT* obj;
    //obj = X509_EXTENSION_get_object(x509_ext);
    //i2a_ASN1_OBJECT(bio, obj);
    if(!X509V3_EXT_print(bio, x509_ext, 0, 4)) {
      //If the extension is not recognized by openssl, such as the voms AC extension,
      //we just output the binary itself
      BIO_printf(bio, "%4s", "");
      M_ASN1_OCTET_STRING_print(bio, x509_ext->value);
    }
    //BIO_write(bio, "\n", 1);

    ext_val.clear();
    for(;;) {
      char s[256];
      int l = BIO_read(bio,s,sizeof(s));
      if(l <= 0) break;
      ext_val.append(s,l);
    }
    BIO_free_all(bio);
 
/*
    X509V3_EXT_METHOD* method;
    STACK_OF(CONF_VALUE)* val;
    CONF_VALUE *nval;
    void *extstr = NULL;
#if (OPENSSL_VERSION_NUMBER >= 0x0090800FL)
    const unsigned char *ext_value_data;
#else
    unsigned char *ext_value_data;
#endif
    ext_value_data = x509_ext->value->data;

    //Get x509 extension method structure
    if (!(method = (X509V3_EXT_METHOD *)(X509V3_EXT_get(x509_ext)))) return false;
    //Decode ASN1 item in data
    if (method->it) {
      //New style ASN1
      extstr = ASN1_item_d2i(NULL, &ext_value_data, x509_ext->value->length,
                                ASN1_ITEM_ptr(method->it));
    }
    else {
      //Old style ASN1
      extstr = method->d2i(NULL, &ext_value_data, x509_ext->value->length);
    }
    
    if((method->i2v) != NULL) {
      val = method->i2v(method, extstr, NULL);
      for (int j = 0; j < sk_CONF_VALUE_num(val); j++) {
        nval = sk_CONF_VALUE_value(val, j);
        std::string name = nval->name;
        std::string val; if(nval->value) val = nval->value;
        if(!val.empty()) ext_val = name + ":" + val + "\n";
        else ext_val = name + "\n";
      }
    }
    else if((method->i2s) != NULL) {
      std::string val = method->i2s(method, extstr);
      if(!val.empty()) ext_val = val;
    }
    else if((method->i2r) != NULL) {
      std::string val = method->i2r(method, extstr);
      if(!val.empty()) ext_val = val;
    }
*/

    return true;
  }

  //subject is expected to be in the format /type0=value0/type1=value1/type2=...
  //where characters may be escaped by '\'
  static X509_NAME* parse_name(char *subject, long chtype, int multirdn) {
    size_t buflen = strlen(subject)+1;
    /* to copy the types and values into. due to escaping, the copy can only become shorter */
    char *buf = (char*)(OPENSSL_malloc(buflen));
    size_t max_ne = buflen / 2 + 1; /* maximum number of name elements */
    char **ne_types =  (char **)(OPENSSL_malloc(max_ne * sizeof (char *)));
    char **ne_values = (char **)(OPENSSL_malloc(max_ne * sizeof (char *)));
    int *mval = (int*)(OPENSSL_malloc (max_ne * sizeof (int)));

    char *sp = subject, *bp = buf;
    int i, ne_num = 0;

    X509_NAME *n = NULL;
    int nid;

    if (!buf || !ne_types || !ne_values) {
      OpenSSLUtilLogger.msg(ERROR, "OpenSSL malloc error");
      goto error;
    }
    if (*subject != '/') {
      OpenSSLUtilLogger.msg(ERROR,"Subject does not start with '/'");
      goto error;
    }
    sp++; // skip leading /

    // no multivalued RDN by default
    mval[ne_num] = 0;
    while (*sp) {
      // collect type
      ne_types[ne_num] = bp;
      while (*sp) {
        if (*sp == '\\') /* is there anything to escape in the type...? */
        {
          if (*++sp) *bp++ = *sp++;
          else {
            OpenSSLUtilLogger.msg(ERROR,"escape character at end of string");
            goto error;
          }
        }
        else if (*sp == '=') {
          sp++;
          *bp++ = '\0';
          break;
        }
        else  *bp++ = *sp++;
      }
      if (!*sp) {
        OpenSSLUtilLogger.msg(ERROR,"end of string encountered while processing type of subject name element #%d",ne_num);
        goto error;
      }
      ne_values[ne_num] = bp;
      while (*sp) {
        if (*sp == '\\') {
          if (*++sp)
            *bp++ = *sp++;
          else {
            OpenSSLUtilLogger.msg(ERROR,"escape character at end of string");
            goto error;
          }
        }
        else if (*sp == '/') {
          sp++;
          // no multivalued RDN by default
          mval[ne_num+1] = 0;
          break;
        }
        else if (*sp == '+' && multirdn) {
          // a not escaped + signals a mutlivalued RDN
          sp++;
          mval[ne_num+1] = -1;
          break;
        }
        else
          *bp++ = *sp++;
      }
      *bp++ = '\0';
      ne_num++;
    }

    if (!(n = X509_NAME_new()))
      goto error;

    for (i = 0; i < ne_num; i++) {
      if ((nid=OBJ_txt2nid(ne_types[i])) == NID_undef) {
        OpenSSLUtilLogger.msg(ERROR,"Subject Attribute %s has no known NID, skipped",ne_types[i]);
        continue;
      }
      if (!*ne_values[i]) {
        OpenSSLUtilLogger.msg(ERROR,"No value provided for Subject Attribute %s skipped",ne_types[i]);
        continue;
      }

      if (!X509_NAME_add_entry_by_NID(n, nid, chtype, (unsigned char*)ne_values[i], -1,-1,mval[i]))
        goto error;
    }

    OPENSSL_free(mval);
    OPENSSL_free(ne_values);
    OPENSSL_free(ne_types);
    OPENSSL_free(buf);
    return n;

error:
    X509_NAME_free(n);
    if (ne_values)
      OPENSSL_free(ne_values);
    if (ne_types)
      OPENSSL_free(ne_types);
    if (buf)
      OPENSSL_free(buf);
    return NULL;
  }

  static int BIO_write_filename_User(BIO *b, const char* file) {
    return BIO_write_filename(b, (char*)file);
  }

  static int BIO_read_filename_User(BIO *b, const char* file) {
    return BIO_read_filename(b, (char*)file);
  }

  static void print_ssl_errors() {
    unsigned long err;
    while((err = ERR_get_error())) {
      OpenSSLUtilLogger.msg(DEBUG,"SSL error: %s, libs: %s, func: %s, reason: %s",
        ERR_error_string(err, NULL),ERR_lib_error_string(err),
        ERR_func_error_string(err),ERR_reason_error_string(err));
    }
  }

#define SERIAL_RAND_BITS        64
  int rand_serial(BIGNUM *b, ASN1_INTEGER *ai) {
    BIGNUM *btmp;
    int ret = 0;
    if (b) btmp = b;
    else btmp = BN_new();
    if (!btmp) return 0;
    if (!BN_pseudo_rand(btmp, SERIAL_RAND_BITS, 0, 0)) goto error;
    if (ai && !BN_to_ASN1_INTEGER(btmp, ai)) goto error;
    ret = 1;
error:
    if (!b) BN_free(btmp);
    return ret;
  }

#undef BSIZE
#define BSIZE 256
  BIGNUM *load_serial(const std::string& serialfile, ASN1_INTEGER **retai) {
    BIO *in=NULL;
    BIGNUM *ret=NULL;
    char buf[1024];
    ASN1_INTEGER *ai=NULL;
    ai=ASN1_INTEGER_new();
    if (ai == NULL) goto err;
    if ((in=BIO_new(BIO_s_file())) == NULL) {
      print_ssl_errors();
      goto err;
    }
    if (BIO_read_filename_User(in,serialfile.c_str()) > 0) {
      if (!a2i_ASN1_INTEGER(in,ai,buf,1024)) {
        OpenSSLUtilLogger.msg(ERROR,"unable to load number from: %s",serialfile.c_str());
        goto err;
      }
      ret=ASN1_INTEGER_to_BN(ai,NULL);
      if (ret == NULL) {
        OpenSSLUtilLogger.msg(ERROR, "error converting number from bin to BIGNUM");
        goto err;
      }
    }

    if (ret && retai) {
      *retai = ai;
      ai = NULL;
    }
err:
    if (in != NULL) BIO_free(in);
    if (ai != NULL) ASN1_INTEGER_free(ai);
      return(ret);
  }

  int save_serial(const std::string& serialfile, char *suffix, BIGNUM *serial, ASN1_INTEGER **retai) {
    char buf[1][BSIZE];
    BIO *out = NULL;
    int ret=0;
    ASN1_INTEGER *ai=NULL;
    int j;

    if (suffix == NULL) j = strlen(serialfile.c_str());
    else j = strlen(serialfile.c_str()) + strlen(suffix) + 1;
    if (j >= BSIZE) {
      OpenSSLUtilLogger.msg(ERROR,"file name too long");
      goto err;
    }

    if (suffix == NULL)
      BUF_strlcpy(buf[0], serialfile.c_str(), BSIZE);
    else {
#ifndef OPENSSL_SYS_VMS
      j = BIO_snprintf(buf[0], sizeof buf[0], "%s.%s", serialfile.c_str(), suffix);
#else
      j = BIO_snprintf(buf[0], sizeof buf[0], "%s-%s", serialfile.c_str(), suffix);
#endif
    }
    out=BIO_new(BIO_s_file());
    if (out == NULL) {
      print_ssl_errors();
      goto err;
    }
    if (BIO_write_filename_User(out,buf[0]) <= 0) {
      perror(serialfile.c_str());
      goto err;
    }
    if ((ai=BN_to_ASN1_INTEGER(serial,NULL)) == NULL) {
      OpenSSLUtilLogger.msg(ERROR,"error converting serial to ASN.1 format");
      goto err;
    }
    i2a_ASN1_INTEGER(out,ai);
    BIO_puts(out,"\n");
    ret=1;
    if (retai) {
      *retai = ai;
      ai = NULL;
    }
err:
    if (out != NULL) BIO_free_all(out);
    if (ai != NULL) ASN1_INTEGER_free(ai);
    return(ret);
  }

#undef POSTFIX
#define POSTFIX ".srl"
  static ASN1_INTEGER *x509_load_serial(const std::string& CAfile, const std::string& serialfile) {
    ASN1_INTEGER *bs = NULL;
    BIGNUM *serial = NULL;

    std::string serial_f;
    if(!serialfile.empty()) serial_f = serialfile;
    else if(!CAfile.empty()){
      std::size_t pos; pos = CAfile.rfind(".");
      if(pos != std::string::npos) serial_f = CAfile.substr(0, pos);
      serial_f.append(".srl");
    }
    else{ return bs;}

    serial = load_serial(serial_f, NULL);
    if (serial == NULL) {
      OpenSSLUtilLogger.msg(ERROR,"load serial from %s failure",serial_f.c_str());
      OpenSSLError();
      return bs;
    }

    if (!BN_add_word(serial,1)) {
      OpenSSLUtilLogger.msg(ERROR,"add_word failure");
      BN_free(serial); return bs;
    }

    if(!save_serial(serial_f, NULL, serial, &bs)) {
      OpenSSLUtilLogger.msg(ERROR,"save serial to %s failure",serial_f.c_str());
      BN_free(serial); return bs;
    }
    BN_free(serial);
    return bs;
  }

#define REQ_SECTION     "req"
#define BITS            "default_bits"
#define KEYFILE         "default_keyfile"
#define DISTINGUISHED_NAME      "distinguished_name"
#define ATTRIBUTES      "attributes"
#define V3_EXTENSIONS   "x509_extensions"
#define REQ_EXTENSIONS  "req_extensions"
#define STRING_MASK     "string_mask"
#define UTF8_IN         "utf8"

#define CA_SECTION      "ca"
#define DEFAULT_CA      "default_ca"
#define ENV_DIR                 "dir"
#define ENV_CERTS               "certs"
#define ENV_CRL_DIR             "crl_dir"
#define ENV_CA_DB               "CA_DB"
#define ENV_NEW_CERTS_DIR       "new_certs_dir"
#define ENV_CERTIFICATE         "certificate"
#define ENV_SERIAL              "serial"
#define ENV_CRLNUMBER           "crlnumber"
#define ENV_CRL                 "crl"
#define ENV_PRIVATE_KEY         "private_key"
#define ENV_RANDFILE            "RANDFILE"
#define ENV_DEFAULT_DAYS        "default_days"
#define ENV_DEFAULT_STARTDATE   "default_startdate"
#define ENV_DEFAULT_ENDDATE     "default_enddate"
#define ENV_DEFAULT_CRL_DAYS    "default_crl_days"
#define ENV_DEFAULT_CRL_HOURS   "default_crl_hours"
#define ENV_DEFAULT_MD          "default_md"
#define ENV_DEFAULT_EMAIL_DN    "email_in_dn"
#define ENV_PRESERVE            "preserve"
#define ENV_POLICY              "policy"
#define ENV_EXTENSIONS          "x509_extensions"
#define ENV_CRLEXT              "crl_extensions"
#define ENV_MSIE_HACK           "msie_hack"
#define ENV_NAMEOPT             "name_opt"
#define ENV_CERTOPT             "cert_opt"
#define ENV_EXTCOPY             "copy_extensions"
#define ENV_UNIQUE_SUBJECT      "unique_subject"
#define ENV_DATABASE            "database"


#define PROXY_EXT_SECTION   "proxy_cert_ext"
#define PROXY_CERTINFO_EXTENSION     "proxyCertInfo"

#if 0
  static bool load_conf_module(CONF* cnf) {
    static int load_config_called = 0;
    if(load_config_called) return true;
    load_config_called = 1;
    if (!cnf) return true;
    OPENSSL_load_builtin_modules();
    if(CONF_modules_load(cnf, NULL, 0) <= 0) {
      OpenSSLUtilLogger.msg(ERROR, "Error configuring OpenSSL");
      OpenSSLError(); return false;
    }
    return true;
  }
#endif

  static CONF* load_conf_file(const std::string& conf_file) {
    CONF* conf = NULL;
    if (!conf_file.empty()) {
      long errorline = -1;
      conf = NCONF_new(NULL);
      if (!NCONF_load(conf, conf_file.c_str(),&errorline)) {
        if (errorline <= 0) {
          OpenSSLUtilLogger.msg(ERROR,"Error when loading the config file: %s", conf_file.c_str());
          return NULL;
        }
        else {
          OpenSSLUtilLogger.msg(ERROR,"Error when loading the config file: %s on line: %ld",
             conf_file.c_str(), errorline);
        }
      }
    }
    return conf;
  }

  static const char* load_conf_test(CONF* conf, const char* group, const char* name) {
    char* exts = NULL;
    exts = NCONF_get_string(conf, group, name);
    if (!exts) OpenSSLError();
    if (exts) {
      //Check syntax of file
      X509V3_CTX ctx;
      X509V3_set_ctx_test(&ctx);
      X509V3_set_nconf(&ctx, conf);
      if(!X509V3_EXT_add_nconf(conf, &ctx, exts, NULL)) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to load conf section %s\n", exts);
        return NULL;
      }
    }
    return exts;
  }

  static const bool load_conf_tox509(CONF* conf, const char* group, const char* name, X509* x509_ca, X509* x509) {
    char* exts = NULL;
    exts = NCONF_get_string(conf, group, name);
    if (!exts) OpenSSLError();
    if (exts) {
      X509V3_CTX ctx;
      X509V3_set_ctx(&ctx, x509_ca, x509, NULL, NULL, 0);
      X509V3_set_nconf(&ctx, conf);
      if(!X509V3_EXT_add_nconf(conf, &ctx, exts, x509)) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to load extension section %s\n to X509 object", exts);
        return false;
      }
    }
    return true;
  }

  static const bool load_conf_tox509req(CONF* conf, const char* group, const char* name, X509_REQ* x509req) {
    char* exts = NULL;
    exts = NCONF_get_string(conf, group, name);
    if (!exts) OpenSSLError();
    if (exts) {
      X509V3_CTX ctx;
      X509V3_set_ctx(&ctx, NULL, NULL, x509req, NULL, 0);
      X509V3_set_nconf(&ctx, conf);
      if(!X509V3_EXT_REQ_add_nconf(conf, &ctx, exts, x509req)) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to load extension section %s\n to X509 request object", exts);
        return false;
      }
    }
    return true;
  }
   
  static char* load_section(CONF* conf, const char* group, const char* name) {
    char* section = NULL;
    section = NCONF_get_string(conf, group, name);
    if(section == NULL) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to find %s from configuration file", name);
    }
    return section;
  }

  static int x509_certify(const EVP_MD *digest, X509 *x, X509 *xca, EVP_PKEY *pkey, 
             const std::string& serialfile, time_t start, time_t lifetime, int clrext, 
             CONF *conf, char *section, ASN1_INTEGER *sno) {
    int ret=0;
    ASN1_INTEGER *bs=NULL;
    EVP_PKEY *upkey;
    std::string serfile;
    std::string x509_sec;
    STACK_OF(CONF_VALUE)* ca_sk = NULL;
    char* ca_sect;

    upkey = X509_get_pubkey(xca);
    EVP_PKEY_copy_parameters(upkey,pkey);
    EVP_PKEY_free(upkey);

    ca_sect = load_section(conf, CA_SECTION, DEFAULT_CA);
    if(ca_sect == NULL) goto end;
    if(!serialfile.empty()) serfile = serialfile;
    else {
      serfile = load_section(conf, ca_sect, ENV_SERIAL);  
      if(serfile.empty()) goto end;
    }
    if(section) x509_sec = section;
    else x509_sec = ENV_EXTENSIONS;

    if (sno) bs = sno;
    else if (!(bs = x509_load_serial("", serfile))) {
      bs = ASN1_INTEGER_new();
      if( bs == NULL || !rand_serial(NULL,bs)) {
        OpenSSLUtilLogger.msg(ERROR,"Out of memory when generate random serial");
        goto end;
      }
    }

    if (!X509_check_private_key(xca,pkey)) {
      OpenSSLUtilLogger.msg(ERROR,"CA certificate and CA private key do not match");
      goto end;
    }
    if (!X509_set_issuer_name(x,X509_get_subject_name(xca))) goto end;
    if (!X509_set_serialNumber(x,bs)) goto end;

    if (X509_gmtime_adj(X509_get_notBefore(x), start) == NULL) goto end;
    if (X509_gmtime_adj(X509_get_notAfter(x), lifetime) == NULL) goto end;

    if (clrext) {
      while (X509_get_ext_count(x) > 0) X509_delete_ext(x, 0);
    }

    X509_set_version(x,2);
    if(conf && !x509_sec.empty()) load_conf_tox509(conf, ca_sect, x509_sec.c_str(), xca, x);
    if (!X509_sign(x,pkey,digest)) goto end;
      ret=1;
end:
    if (!ret)
      ERR_clear_error();
    if (!sno) ASN1_INTEGER_free(bs);

    return ret;
  }

  static ASN1_UTCTIME* utc_to_asn1time(const AuthN::Utils::Time& t) {
    std::string t_str = t.str(AuthN::Utils::MDSTime);
    if(t_str.length() < 2) return NULL;
    ASN1_UTCTIME* s = ASN1_UTCTIME_new();
    if(!s) return NULL;
    if(ASN1_UTCTIME_set_string(s,(char*)(t_str.c_str() +2))) return s;
    ASN1_UTCTIME_free(s);
    return NULL;
  }

  static ASN1_INTEGER* long_to_asn1int(const long val) {
    ASN1_INTEGER* asn1=NULL;
    if (!(asn1 = ASN1_INTEGER_new())) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to allocate ASN1_INTEGER memory" ); return NULL; 
    }
    if(!ASN1_INTEGER_set(asn1, val)) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to set ASN1_INTEGER object"); ASN1_INTEGER_free(asn1); return NULL; 
    }
    return asn1;
  }

  static bool parse_req_conf(X509_REQ* req, CONF* conf, bool attribs, unsigned long chtype) {
    char *p,*q;
    char* type;
    CONF_VALUE* v;
    STACK_OF(CONF_VALUE)* dn_sk = NULL;
    STACK_OF(CONF_VALUE)* attr_sk = NULL;
    char *dn_sect, *attr_sect;
    X509_NAME* subj;
    subj = X509_REQ_get_subject_name(req);

    dn_sect = NCONF_get_string(conf, REQ_SECTION, DISTINGUISHED_NAME);
    if(dn_sect == NULL) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to find %s from configuration file", DISTINGUISHED_NAME);
      return false;
    }
    dn_sk = NCONF_get_section(conf,dn_sect);
    if(dn_sk == NULL) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to get %s section", dn_sect);
      return false;
    }
    for (int i = 0; i < sk_CONF_VALUE_num(dn_sk); i++) {
      int mval;
      v = sk_CONF_VALUE_value(dn_sk,i);
      p = q = NULL;
      type=v->name;
      for(p = v->name; *p ; p++)
#ifndef CHARSET_EBCDIC
        if ((*p == ':') || (*p == ',') || (*p == '.')) {
#else
        if ((*p == os_toascii[':']) || (*p == os_toascii[',']) || (*p == os_toascii['.'])) {
#endif
          p++; if(*p) type = p; break;
        }
#ifndef CHARSET_EBCDIC
      if (*p == '+')
#else
      if (*p == os_toascii['+'])
#endif
      {
        p++; mval = -1;
      }
      else mval = 0;
      if (!X509_NAME_add_entry_by_txt(subj,type, chtype, (unsigned char *) v->value,-1,-1,mval)) { OpenSSLError(); return false; }
    }
    if(!X509_NAME_entry_count(subj)) {
      OpenSSLUtilLogger.msg(ERROR, "No objects specified in config file");
      return false;
    }

    if (attribs) {
      attr_sect = NCONF_get_string(conf, REQ_SECTION, ATTRIBUTES);
      if(attr_sect == NULL) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to find %s from configuration file", ATTRIBUTES);
        return false;
      }
      attr_sk = NCONF_get_section(conf, attr_sect);
      if(attr_sk == NULL) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to get %s section", attr_sect);
        return false;
      }
      for(int i = 0; i < sk_CONF_VALUE_num(attr_sk); i++) {
        v = sk_CONF_VALUE_value(attr_sk,i);
        if(!X509_REQ_add1_attr_by_txt(req, v->name, chtype, (unsigned char *)v->value, -1)) { OpenSSLError(); return false; }
      }
    }
    return true;
  }

  static const EVP_MD* get_digest(Logger& logger, const std::string& digest_name, const EVP_PKEY* privkey) {
    const EVP_MD* digest = NULL;
    int digest_nid;
    char* digest_str = NULL;

    //If private key is available, try to get the digest information from it
    //Because EVP_PKEY_get_default_digest_nid is only available on openssl version newer than 1.0.0,
    //for relatively older version of openssl, digest info of private key will not be obtained.
#if OPENSSL_VERSION_NUMBER >= 0x10000000L
    int i = EVP_PKEY_get_default_digest_nid((EVP_PKEY*)privkey, &digest_nid);
    if(privkey != NULL) { 
      if(i <= 0) {
        logger.msg(INFO, "There is no digest in issuer's private key object");
      }
      else if(i == 2) { //return 2 if the digest is mandatory (other digests can not be used)
        digest_str = (char *)OBJ_nid2sn(digest_nid);
      }
      else { //returns 1 if the message digest is advisory (that is other digests can be used) 
        if(digest_name.empty()) digest_str = (char *)OBJ_nid2sn(digest_nid);
      }
    }
#endif
    if(digest_str == NULL && !digest_name.empty())  digest_str = (char*)(digest_name.c_str());
    if(digest_str != NULL)
      if((digest = EVP_get_digestbyname(digest_str)) == NULL)
        logger.msg(INFO, "%s is an unsupported digest type", digest_str);

    if(digest == NULL) digest = EVP_sha1();

    // Check whether the digest algorithm is SHA1 or SHA2
    // For openssl version older than 0.9.8f, SHA2 is not supported
    digest_nid = EVP_MD_type(digest);
#if OPENSSL_VERSION_NUMBER >= 0x0090800fL
    if((digest_nid != NID_sha1) && (digest_nid != NID_sha224) && (digest_nid != NID_sha256) && (digest_nid != NID_sha384) && (digest_nid != NID_sha512)) {
      logger.msg(ERROR, "The signing algorithm %s is not allowed,it should be SHA1 or SHA2 to sign certificate requests",
      OBJ_nid2sn(digest_nid));
      return NULL;
    }
#else
    if(digest_nid != NID_sha1) {
      logger.msg(ERROR, "The signing algorithm %s is not allowed,it should be SHA1 to sign certificate requests",
      OBJ_nid2sn(md_nid));
      return NULL;
    }
#endif

    logger.msg(INFO, "The digest that is used is: %s", OBJ_nid2sn(digest_nid));

    return digest;
  } 

  static const EVP_MD* get_digest_from_conf(Logger& logger, const std::string& conf_file, const EVP_PKEY* privkey) { 
    const EVP_MD* digest = NULL;
    CONF* conf = NULL;
    conf = load_conf_file(conf_file);

    char* ca_sect;
    char* md_str;
    ca_sect = load_section(conf, CA_SECTION, DEFAULT_CA);
    if(ca_sect != NULL) {
      md_str = load_section(conf, ca_sect, ENV_DEFAULT_MD);
    }
    if (md_str == NULL) ERR_clear_error();
    if (md_str != NULL) {
      digest = get_digest(logger, md_str, privkey);
    }
    NCONF_free(conf);

    return digest;
  }

  bool make_request(X509_REQ* req, const EVP_PKEY* pubkey, const EVP_PKEY* privkey, bool req_attribs, const std::string& conf_file, const std::string& subject_name) {
    CONF* req_conf = NULL;
    req_conf = load_conf_file(conf_file);

    const char* req_ext = load_conf_test(req_conf, REQ_SECTION, REQ_EXTENSIONS);
    if(!req_ext) { NCONF_free(req_conf); return false; }

    load_conf_tox509req(req_conf, REQ_SECTION, REQ_EXTENSIONS, req);

    char* dn_sect = NCONF_get_string(req_conf, REQ_SECTION, DISTINGUISHED_NAME);
    if (dn_sect == NULL) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to find %s from configuration file", DISTINGUISHED_NAME);
      NCONF_free(req_conf); return false;
    }
    STACK_OF(CONF_VALUE)* dn_sk = NULL; 
    dn_sk = NCONF_get_section(req_conf,dn_sect);
    if(dn_sk == NULL) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to get %s section", dn_sect);
      NCONF_free(req_conf); return false;
    }
    char* attr_sect = NCONF_get_string(req_conf, REQ_SECTION, ATTRIBUTES);
    if(attr_sect == NULL) {
      OpenSSLUtilLogger.msg(WARNING, "Failed to find %s from configuration file", ATTRIBUTES);
      ERR_clear_error();
    }
    else {
      STACK_OF(CONF_VALUE)* attr_sk = NULL;
      attr_sk=NCONF_get_section(req_conf,attr_sect);
      if(attr_sk == NULL) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to get %s section", attr_sect);
        NCONF_free(req_conf); return false;
      }
    }      
    X509_REQ_set_version(req, 0L); //version 1
  
    unsigned long chtype = MBSTRING_ASC;  
    if(!subject_name.empty()) {
      X509_NAME *subject = NULL;
      subject = parse_name((char*)(subject_name.c_str()), chtype, 0);
      X509_REQ_set_subject_name(req, subject);
      X509_NAME_free(subject);
    }
    else {
      parse_req_conf(req, req_conf, req_attribs, chtype);
    }
    NCONF_free(req_conf);

    const EVP_MD* digest;
    digest = get_digest_from_conf(OpenSSLUtilLogger, conf_file, privkey);

    X509_REQ_set_pubkey(req, (EVP_PKEY*)pubkey);
    if(!X509_REQ_sign(req, (EVP_PKEY*)privkey, digest)) { OpenSSLError(); return false; }
    return true;
  }

  static bool copy_extensions(X509* x, const X509_REQ* req, bool override) {
    STACK_OF(X509_EXTENSION)* exts = NULL;
    X509_EXTENSION *ext, *tmpext;
    ASN1_OBJECT *obj;
    int i, idx; 
    if (!x || !req) return true;
    exts = X509_REQ_get_extensions((X509_REQ*)req);

    for(i = 0; i < sk_X509_EXTENSION_num(exts); i++) {
      ext = sk_X509_EXTENSION_value(exts, i);
      obj = X509_EXTENSION_get_object(ext);
      idx = X509_get_ext_by_OBJ(x, obj, -1);
      if(idx != -1) {
        //If not override existing extension
        if(!override) continue;
        //Delete all extensions of same type
        do {
          tmpext = X509_get_ext(x, idx);
          X509_delete_ext(x, idx);
          X509_EXTENSION_free(tmpext);
          idx = X509_get_ext_by_OBJ(x, obj, -1);
        } while (idx != -1);
      }
      if (!X509_add_ext(x, ext, -1)) return false;;
    }
    return true;
  }

  bool selfsign_request(X509** cert, const X509_REQ* req, const EVP_PKEY* privkey,  const time_t from, const time_t till, const std::string& dn_name, const long serial, const std::string& conf_file) {
    bool ret = false;
    *cert = X509_new(); 
    X509_NAME *name = NULL;
    unsigned long chtype = MBSTRING_ASC;
    if(!dn_name.empty()) {
      name = parse_name((char*)(dn_name.c_str()), chtype, 0);
      X509_set_subject_name(*cert, name);
      X509_set_issuer_name(*cert, name);
      X509_NAME_free(name);
    }
    else {
      X509_set_subject_name(*cert, X509_REQ_get_subject_name(req));
      X509_set_issuer_name(*cert, X509_REQ_get_subject_name(req));
    }

    CONF* req_conf = NULL;
    req_conf = load_conf_file(conf_file);
    const char* v3_ext = load_conf_test(req_conf, REQ_SECTION, V3_EXTENSIONS);
    if(!v3_ext) { NCONF_free(req_conf); X509_free(*cert); return false; }

    //Set version to V3
    if((v3_ext != NULL) && !X509_set_version(*cert, 2)) { 
      OpenSSLUtilLogger.msg(ERROR, "v3 extension is provided in the req section of configuration file, failed to set the V3 version for the X509 object");
      NCONF_free(req_conf); X509_free(*cert); return false; 
    }
    ASN1_INTEGER* asn1 = NULL; 
    if((serial != -1) && ((asn1 = long_to_asn1int(serial)) != NULL)) { 
      X509_set_serialNumber(*cert, asn1); 
      ASN1_INTEGER_free(asn1); 
    }
    else if(!rand_serial(NULL, X509_get_serialNumber(*cert)));
    AuthN::Utils::Time s = from;
    ASN1_UTCTIME* not_before = utc_to_asn1time(s);
    AuthN::Utils::Time e = till;
    ASN1_UTCTIME* not_after = utc_to_asn1time(e);
    if((!not_before) || (!not_after)) {
      NCONF_free(req_conf); X509_free(*cert);
      if(not_before) ASN1_UTCTIME_free(not_before);
      if(not_after) ASN1_UTCTIME_free(not_after);
      return false;
    }
    X509_set_notBefore(*cert, not_before);
    X509_set_notAfter(*cert, not_after);
    ASN1_UTCTIME_free(not_before);
    ASN1_UTCTIME_free(not_after);

    EVP_PKEY *tmppkey;
    tmppkey = X509_REQ_get_pubkey((X509_REQ*)req);
    if (!tmppkey) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to load public key from X509_REQ object");
      NCONF_free(req_conf); X509_free(*cert); return false;
    } 
    X509_set_pubkey(*cert, tmppkey);
    EVP_PKEY_free(tmppkey);

    //Add v3 extensions
    if(v3_ext != NULL) load_conf_tox509(req_conf, REQ_SECTION, V3_EXTENSIONS, *cert, *cert);

    NCONF_free(req_conf);

    //Load digest from configuration
    const EVP_MD* digest;
    digest = get_digest_from_conf(OpenSSLUtilLogger, conf_file, privkey);

    if(digest == NULL) digest = EVP_sha1();
#ifndef OPENSSL_NO_DSA
    if (privkey->type == EVP_PKEY_DSA)
      digest=EVP_dss1();
#endif
#ifndef OPENSSL_NO_ECDSA
    if (privkey->type == EVP_PKEY_EC)
      digest = EVP_ecdsa();
#endif
    if (!(X509_sign(*cert, (EVP_PKEY*)privkey,digest))) {
      X509_free(*cert);
      OpenSSLError(); return false;
    }
    return true;
  }

  bool sign_request(X509** cert, const X509_REQ* req, const EVP_PKEY* issuer_privkey, const X509* issuer_cert, const time_t from, const time_t till, const std::string& dn_name, const std::string& serial_file, const std::string& conf_file) {
    if(issuer_privkey == NULL) { OpenSSLUtilLogger.msg(ERROR, "The private key for signing is not initialized"); return false; }
    bool res = false;
    *cert = X509_new();
    if(req == NULL) { OpenSSLUtilLogger.msg(ERROR, "X509_REQ is not provided"); return false; }

    EVP_PKEY* req_pubkey = NULL;
    req_pubkey = X509_REQ_get_pubkey((X509_REQ*)req);
    if(!req_pubkey) { OpenSSLUtilLogger.msg(ERROR, "Error when extracting public key from request");
      OpenSSLError(); return false;
    }
    if(!X509_REQ_verify((X509_REQ*)req, req_pubkey)){
      OpenSSLUtilLogger.msg(ERROR,"Failed to verify the request");
      EVP_PKEY_free(req_pubkey); OpenSSLError(); return false;
    }
    if(*cert == NULL) *cert = X509_new();
    X509_set_pubkey(*cert, req_pubkey);
    EVP_PKEY_free(req_pubkey);

    X509_NAME *subject = NULL;
    unsigned long chtype = MBSTRING_ASC;  //TODO
    if(!dn_name.empty()) {
      subject = parse_name((char*)(dn_name.c_str()), chtype, 0);
      X509_set_subject_name(*cert, subject);
      X509_NAME_free(subject);
    }
    else 
      X509_set_subject_name(*cert, req->req_info->subject);

    const EVP_MD* digest;
    digest = get_digest_from_conf(OpenSSLUtilLogger, conf_file, issuer_privkey);

    CONF* extconf = NULL;
    extconf = load_conf_file(conf_file);

    //Add extensions to certificate object
    copy_extensions(*cert, req, true);

    time_t lifetime = till - from;
    AuthN::Utils::Time t1 = from; 
    AuthN::Utils::Time t2;
    time_t start;
    if(t1 > t2) start = t1.GetTime() - t2.GetTime();
    else start = 0;

    if (!x509_certify(digest, *cert, (X509*)issuer_cert,
                      (EVP_PKEY* )issuer_privkey, serial_file, start, lifetime, 0,
                      extconf, NULL, NULL)) {
      OpenSSLUtilLogger.msg(ERROR,"Can not sign a EEC"); OpenSSLError();
    }

    if(extconf) NCONF_free(extconf);

    return true;
  }

  static bool str_to_asn1int(char *value, ASN1_INTEGER **aint) {
    ASN1_INTEGER *itmp;
    if(!(itmp = s2i_ASN1_INTEGER(NULL, value))) return false;
    *aint = itmp;
    return true;
  }

  bool sign_proxy_request(X509** cert, const X509_REQ* req, const EVP_PKEY* issuer_privkey, const X509* issuer_cert, const time_t from, const time_t till, const std::string& proxypolicy, const std::string& conf_file){
    EVP_PKEY* req_pubkey = NULL;
    ASN1_INTEGER* serial_number = NULL;
    ASN1_OBJECT* pl_obj= NULL;
    ASN1_OCTET_STRING* policy_string = NULL;
    ASN1_INTEGER* policy_pathlen = NULL;
    X509_NAME* subject_name = NULL;
    X509_NAME_ENTRY* name_entry = NULL;
    char* CN_name = NULL;
    const EVP_MD *digest = EVP_sha1();
    time_t validity_start = time(NULL);
    time_t validity_end = (time_t)(-1);
    const char* need_ext = NULL;
    X509_EXTENSION* ext = NULL;
    STACK_OF(X509_EXTENSION)* req_exts = NULL;
    X509_EXTENSION* proxy_ext = NULL;
    int num, i, nid;
    CONF* proxy_conf = NULL;
    const char* conf_pci = NULL;
    std::string conf_pci_str;
    bool conf_pci_critical = false;
    std::string conf_pci_language;
    std::string conf_pci_pathlen;
    std::string conf_pci_policy;
    std::size_t pos1, pos2;
    bool ret = false;

    *cert = X509_new();
    req_pubkey = X509_REQ_get_pubkey((X509_REQ*)req);
    if(!req_pubkey) {
      OpenSSLUtilLogger.msg(ERROR, "Error when extracting public key from request");
      OpenSSLError(); return false;
    }
    if(!X509_REQ_verify((X509_REQ*)req, req_pubkey)){
      OpenSSLUtilLogger.msg(ERROR,"Failed to verify the request"); OpenSSLError(); goto err;
    }

    req_exts = X509_REQ_get_extensions((X509_REQ*)req);
    num = sk_X509_EXTENSION_num(req_exts);
    for(i = 0; i < num; i++) {
      ext = sk_X509_EXTENSION_value(req_exts, i);
      nid = OBJ_obj2nid(X509_EXTENSION_get_object(ext));
      if (nid == NID_proxyCertInfo) { 
        proxy_ext = ext; break;
      }
    }

    serial_number = ASN1_INTEGER_new();
    if(!serial_number) goto err;
    if(!rand_serial(NULL, serial_number)) goto err;
    if (!X509_set_serialNumber(*cert,serial_number)) goto err;
    ASN1_INTEGER_free(serial_number); serial_number=NULL;
    X509_set_version(*cert,2L);

    /*
    Proxy certificates do not need KeyUsage extension. But
    some old software still expects it to be present.

    From RFC3820:
    If the Proxy Issuer certificate has the KeyUsage extension, the
    Digital Signature bit MUST be asserted.
    */
    need_ext = "critical,digitalSignature,keyEncipherment";
    ext = X509V3_EXT_conf_nid(NULL, NULL, NID_key_usage, (char*)need_ext);
    if(ext != NULL) { X509_add_ext(*cert, ext, -1); X509_EXTENSION_free(ext); }

    /*
    From RFC3820:
    If a certificate is a Proxy Certificate, then the proxyCertInfo
    extension MUST be present, and this extension MUST be marked as
    critical.
  
    The pCPathLenConstraint field, if present, specifies the maximum
    depth of the path of Proxy Certificates that can be signed by this
    Proxy Certificate. 

    The proxyPolicy field specifies a policy on the use of this
    certificate for the purposes of authorization.  Within the
    proxyPolicy, the policy field is an expression of policy, and the
    policyLanguage field indicates the language in which the policy is
    expressed.

    *  id-ppl-inheritAll indicates that this is an unrestricted proxy
       that inherits all rights from the issuing PI.  An unrestricted
       proxy is a statement that the Proxy Issuer wishes to delegate all
       of its authority to the bearer (i.e., to anyone who has that proxy
       certificate and can prove possession of the associated private
       key).  For purposes of authorization, this an unrestricted proxy
       effectively impersonates the issuing PI.

    *  id-ppl-independent indicates that this is an independent proxy
       that inherits no rights from the issuing PI.  This PC MUST be
       treated as an independent identity by relying parties.  The only
       rights this PC has are those granted explicitly to it.
    */

    proxy_conf = load_conf_file(conf_file);
    conf_pci = NCONF_get_string(proxy_conf, PROXY_EXT_SECTION, PROXY_CERTINFO_EXTENSION);
    if (conf_pci == NULL) {
      OpenSSLUtilLogger.msg(INFO, "Cannot find %s from configuration file", PROXY_CERTINFO_EXTENSION);
    }
    NCONF_free(proxy_conf);
    conf_pci_str = conf_pci;
    if(!conf_pci_str.empty()) {
      if(conf_pci_str.find("critical") != std::string::npos) conf_pci_critical = true;
      pos1 = conf_pci_str.find("language");
      if(pos1 != std::string::npos) {
        pos1 = conf_pci_str.find(":", pos1);
        if(pos1 != std::string::npos) {
          pos2 = conf_pci_str.find(",", pos1);
          if(pos2 != std::string::npos)
            conf_pci_language = conf_pci_str.substr(pos1+1, pos2 - pos1 -1);
          else
            conf_pci_language = conf_pci_str.substr(pos1+1);
        }
      }
      pos1 = conf_pci_str.find("pathlen");
      if(pos1 != std::string::npos) {
        pos1 = conf_pci_str.find(":", pos1);
        if(pos1 != std::string::npos) {
          pos2 = conf_pci_str.find(",", pos1);
          if(pos2 != std::string::npos)
            conf_pci_pathlen = conf_pci_str.substr(pos1+1, pos2 - pos1 -1);
          else
            conf_pci_pathlen = conf_pci_str.substr(pos1+1);
        }
      }
      pos1 = conf_pci_str.find("policy");
      if(pos1 != std::string::npos) {
        pos1 = conf_pci_str.find(":", pos1);
        if(pos1 != std::string::npos) {
          pos2 = conf_pci_str.find(",", pos1);
          if(pos2 != std::string::npos)
            conf_pci_policy = conf_pci_str.substr(pos1+1, pos2 - pos1 -1);
          else
            conf_pci_policy = conf_pci_str.substr(pos1+1);
        }
      }
    }

    PROXY_CERT_INFO_EXTENSION proxy_info;
    PROXY_POLICY proxy_policy;
    memset(&proxy_info,0,sizeof(proxy_info));
    memset(&proxy_policy,0,sizeof(proxy_policy));
    proxy_info.pcPathLengthConstraint=NULL;
    proxy_info.proxyPolicy=&proxy_policy;
    proxy_policy.policyLanguage=NULL;
    proxy_policy.policy=NULL;

    if(!proxypolicy.empty()) {
      pl_obj=OBJ_nid2obj(NID_id_ppl_anyLanguage);  // Proxy with policy
      if(!pl_obj) goto err;
      policy_string=ASN1_OCTET_STRING_new();
      if(!policy_string) goto err;
      ASN1_OCTET_STRING_set(policy_string,(const unsigned char*)(proxypolicy.c_str()),proxypolicy.length());
      proxy_policy.policyLanguage=pl_obj;
      proxy_policy.policy=policy_string;
      if(X509_add1_ext_i2d(*cert,NID_proxyCertInfo,&proxy_info,1,X509V3_ADD_REPLACE) != 1) goto err;
    } 
    else if(proxy_ext != NULL) {
      X509_add_ext(*cert, proxy_ext, -1);
    } 
    else if(!conf_pci_language.empty()) {
      pl_obj=OBJ_txt2obj(conf_pci_language.c_str(),0);
      if(!pl_obj) goto err;
      policy_string=ASN1_OCTET_STRING_new();
      if(!policy_string) goto err;
      str_to_asn1int((char*)(conf_pci_pathlen.c_str()), &policy_pathlen);
      proxy_info.pcPathLengthConstraint = policy_pathlen;
      ASN1_OCTET_STRING_set(policy_string,(const unsigned char*)(conf_pci_policy.c_str()),conf_pci_policy.length());
      proxy_policy.policyLanguage=pl_obj;
      proxy_policy.policy=policy_string;
      if(X509_add1_ext_i2d(*cert,NID_proxyCertInfo,&proxy_info,1,X509V3_ADD_REPLACE) != 1) goto err;
    }
    else{
      pl_obj=OBJ_nid2obj(NID_id_ppl_inheritAll);  // Unrestricted proxy
      if(!pl_obj) goto err;
      proxy_policy.policyLanguage=pl_obj;
      if(X509_add1_ext_i2d(*cert,NID_proxyCertInfo,&proxy_info,1,X509V3_ADD_REPLACE) != 1) goto err;
    }

    if(pl_obj) { 
      if(X509_add1_ext_i2d(*cert,NID_proxyCertInfo,&proxy_info,1,X509V3_ADD_REPLACE) != 1) goto err; 
    }
    if(policy_string) ASN1_OCTET_STRING_free(policy_string); policy_string=NULL;
    if(pl_obj)ASN1_OBJECT_free(pl_obj); pl_obj=NULL;

    subject_name=X509_get_subject_name((X509*)issuer_cert); //subject name of the issuer
    if(!subject_name) goto err;
    subject_name=X509_NAME_dup(subject_name);
    if(!subject_name) goto err;
    if(!X509_set_issuer_name(*cert,subject_name)) goto err;

    unsigned char md[SHA_DIGEST_LENGTH];
    long  sub_hash;
    unsigned int len;
#if (OPENSSL_VERSION_NUMBER >= 0x0090800fL)
    ASN1_digest((int (*)(void*, unsigned char**))i2d_PUBKEY, EVP_sha1(), (char*)req_pubkey,md,&len);
#else
    ASN1_digest((int(*)())i2d_PUBKEY, EVP_sha1(), (char*)req_pubkey,md,&len);
#endif
    sub_hash = md[0] + (md[1] + (md[2] + (md[3] >> 1) * 256) * 256) * 256;
    CN_name = (char*)malloc(sizeof(long)*4 + 1);
    snprintf(CN_name, sizeof(long)*4 + 1, "%ld", sub_hash);

    if((name_entry = X509_NAME_ENTRY_create_by_NID(&name_entry, NID_commonName, V_ASN1_APP_CHOOSE,
                                     (unsigned char *) CN_name, -1)) == NULL) { 
      OpenSSLUtilLogger.msg(ERROR, "Can not create name entry CN for proxy certificate");goto err;
    }

    if(!X509_NAME_add_entry(subject_name, name_entry, X509_NAME_entry_count(subject_name), 0) ||
       !X509_set_subject_name(*cert, subject_name)) {
      OpenSSLUtilLogger.msg(ERROR, "Can not set CN in proxy certificate"); goto err;
    }

    free(CN_name); CN_name = NULL;
    X509_NAME_free(subject_name); subject_name = NULL;
    X509_NAME_ENTRY_free(name_entry); name_entry = NULL;

    validity_start = from;
    validity_end = till;
    //Set "notBefore"
    if( X509_cmp_time(X509_get_notBefore((X509*)issuer_cert), &validity_start) < 0) {
      X509_time_adj(X509_get_notBefore(*cert), 0L, &validity_start);
    }
    else {
      X509_set_notBefore(*cert, X509_get_notBefore((X509*)issuer_cert));
    }
    //Set "not After"
    if(validity_end == (time_t)(-1)) {
      X509_set_notAfter(*cert,X509_get_notAfter((X509*)issuer_cert));
    } else {
      X509_gmtime_adj(X509_get_notAfter(*cert), (validity_end-validity_start));
    }
    X509_set_pubkey(*cert,req_pubkey);
    EVP_PKEY_free(req_pubkey); req_pubkey=NULL;

    if(!X509_sign(*cert,(EVP_PKEY*)issuer_privkey,digest)) {
      OpenSSLUtilLogger.msg(ERROR, "Failed to sign the proxy certificate"); OpenSSLError(); goto err;
    }
    else OpenSSLUtilLogger.msg(INFO, "Succeeded to sign the proxy certificate");

    ret = true;

err:
    OpenSSLError();
    if(req_pubkey) EVP_PKEY_free(req_pubkey);
    if(serial_number) ASN1_INTEGER_free(serial_number);
    if(pl_obj) ASN1_OBJECT_free(pl_obj);
    if(policy_string) ASN1_OCTET_STRING_free(policy_string);
    if(policy_pathlen) { ASN1_INTEGER_free(policy_pathlen); }
    if(subject_name) X509_NAME_free(subject_name);
    if(CN_name)free(CN_name); CN_name = NULL;
    if(name_entry)X509_NAME_ENTRY_free(name_entry);
    return ret;
  }

  X509_STORE* setup_verify(const std::string& CAfile, const std::string& CApath) {
    X509_STORE* store;
    X509_LOOKUP* lookup;
    if(!(store = X509_STORE_new())) goto end;
    lookup=X509_STORE_add_lookup(store,X509_LOOKUP_file());
    if(lookup == NULL) goto end;
    if(!CAfile.empty()) {
      if(!X509_LOOKUP_load_file(lookup,CAfile.c_str(),X509_FILETYPE_PEM)) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to load CA file %s", CAfile.c_str());
        goto end;
      }
    } else X509_LOOKUP_load_file(lookup,NULL,X509_FILETYPE_DEFAULT);

    lookup=X509_STORE_add_lookup(store,X509_LOOKUP_hash_dir());
    if(lookup == NULL) goto end;
    if(!CApath.empty()) {
      if(!X509_LOOKUP_add_dir(lookup,CApath.c_str(),X509_FILETYPE_PEM)) {
        OpenSSLUtilLogger.msg(ERROR, "Failed to load CA directory %s", CApath.c_str());
        goto end;
      }
    } else X509_LOOKUP_add_dir(lookup,NULL,X509_FILETYPE_DEFAULT);

    ERR_clear_error();
    return store;
    end:
    X509_STORE_free(store);
    return NULL;
  }

  // Usint pthreads as sufficiently portable interface.
  pthread_mutex_t * OpenSSLInitializer::ssl_locks = NULL;
  int OpenSSLInitializer::ssl_locks_num = 0;

  OpenSSLInitializer::OpenSSLInitializer(void) {
    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
    static bool initialized = false;
    pthread_mutex_lock(&lock);
    if(initialized) return;

    SSL_load_error_strings();
    SSL_library_init();

    // We could RAND_seed() here. But since 0.9.7 OpenSSL
    // knows how to make use of OS specific source of random
    // data. I think it's better to let OpenSSL do a job.

    // Always make sure our own locks are installed
    int num = CRYPTO_num_locks();
    if(num > 0) {
      ssl_locks = new pthread_mutex_t[num];
      for(int n = 0; n<num; ++n) pthread_mutex_init(ssl_locks+n,NULL);
      ssl_locks_num = num;
      CRYPTO_set_locking_callback(&ssl_locking_cb);
      CRYPTO_set_id_callback(&ssl_id_cb);
    }
    OpenSSL_add_all_algorithms();

    initialized=true;

    pthread_mutex_unlock(&lock);
  }

  unsigned long OpenSSLInitializer::ssl_id_cb(void) {
    return (unsigned long)pthread_self();
  }

  void OpenSSLInitializer::ssl_locking_cb(int mode, int n, const char * /*s_*/, int /*n_*/){
    if(!ssl_locks) return; // Non-threaded environment? Or maybe _exit(-1) ?
    if((n < 0) || (n >= ssl_locks_num)) _exit(-1); // fatal
    if(mode & CRYPTO_LOCK) {
      pthread_mutex_lock(ssl_locks+n);
    } else {
      pthread_mutex_unlock(ssl_locks+n);
    };
  }

  void OpenSSLInitializer::ThreadCleanup(void) {
    ERR_remove_state(0);
  }

  static OpenSSLInitializer openssl_initializer;

}
}

