#include <pthread.h>
#include <list>
#include <string>
#include <openssl/asn1.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pkcs12.h>
#include <openssl/err.h>

#include "datetime.h"

namespace AuthN {
namespace OpenSSL {

  typedef enum {CRED_PEM, CRED_DER, CRED_PKCS, CRED_UNKNOWN} Credformat;

  Credformat getFormat(BIO * in, const bool is_file = true);

  void loadCertificateChain(BIO* certchainbio, STACK_OF(X509) **certchain);
  void loadCertificate(BIO* bio, X509* &x509, STACK_OF(X509)** certchain, const bool is_file=true);
  void loadCertificateFile(const std::string& certfile, X509* &x509, STACK_OF(X509) **certchain);
  void loadKey(BIO* bio, EVP_PKEY* &pkey, const std::string& passphrase = "", const std::string& prompt_info = "", const bool is_file = true);
  void loadKeyFile(const std::string& keyfile, EVP_PKEY* &pkey, const std::string& passphrase = "");
  void loadCRLFile(const std::string& crlfile, X509_CRL* &crl, bool is_pem = true);

  class AutoBIO {
   private:
    BIO* bio_;
   public:
    AutoBIO(BIO* bio):bio_(bio) { };
    ~AutoBIO(void) { if(bio_) { (BIO_set_close(bio_,BIO_CLOSE) == 1); BIO_free_all(bio_); } };
    operator bool(void) { return (bio_ != NULL); };
    operator BIO*(void) { return bio_; };
    BIO& operator*(void) const { return *bio_; };
    BIO* operator->(void) const { return bio_; };
  };

  class AutoX509 {
   private:
    X509* x509_;
   public:
    AutoX509(X509* x509):x509_(x509) { };
    ~AutoX509(void) { if(x509_) { X509_free(x509_); } };
    operator bool(void) { return (x509_ != NULL); };
    operator X509*(void) { return x509_; };
    X509& operator*(void) const { return *x509_; };
    X509* operator->(void) const { return x509_; };
  };
 
  AuthN::Utils::Time asn1_to_utctime(const ASN1_UTCTIME *s);

#define PASS_MIN_LENGTH 4
#define PASS_MAX_LENGTH 20
  typedef struct pw_cb_data {
    const void *password;
    const char *prompt_info;
  } PW_CB_DATA;
  int passwordcb(char *buf, int bufsiz, int verify, void *cb_tmp);

  typedef struct { std::string subject_name; std::string policy; } ProxyPolicy;
  typedef std::list<ProxyPolicy> ProxyPolicyList;
  typedef struct { 
    std::string ca_path;
    ProxyPolicyList ppl;
    int validation_mode;
  } ValidationContext;

  int get_proxy_auth_ex_data_idx(void);
  int verify_callback(int ok, X509_STORE_CTX *ctx);

  enum KeyUsageType {
    // KeyUsage
    DigitalSignature,    ///Certificate can be used to create digital signatures
    NonRepudiation,      ///Certificate can be used for non-repudiation
    KeyEncipherment,     ///Certificate can be used for encrypting / decrypting keys
    DataEncipherment,    ///Certificate can be used for encrypting / decrypting data
    KeyAgreement,        ///Certificate can be used for key agreement
    KeyCertificateSign,  ///Certificate can be used for key certificate signing
    CRLSign,             ///Certificate can be used to sign Certificate Revocation Lists
    EncipherOnly,        ///Certificate can only be used for encryption
    DecipherOnly,        ///Certificate can only be used for decryption

    // ExtKeyUsage
    ServerAuth,       ///ertificate can be used for server authentication (e.g. web server), id = "1.3.6.1.5.5.7.3.1". 
    ClientAuth,       ///Certificate can be used for client authentication (e.g. web browser), id = "1.3.6.1.5.5.7.3.2". 
    CodeSigning,      ///Certificate can be used to sign code, id = "1.3.6.1.5.5.7.3.3". 
    EmailProtection,  ///Certificate can be used to sign / encrypt email, id = "1.3.6.1.5.5.7.3.4".
    IPSecEndSystem,   ///Certificate can be used to authenticate a endpoint in IPSEC, id = "1.3.6.1.5.5.7.3.5".
    IPSecTunnel,      ///Certificate can be used to authenticate a tunnel in IPSEC, id = "1.3.6.1.5.5.7.3.6".
    IPSecUser,        ///Certificate can be used to authenticate a user in IPSEC, id = "1.3.6.1.5.5.7.3.7".
    TimeStamping,     ///Certificate can be used to create a "time stamp" signature, id = "1.3.6.1.5.5.7.3.8".
    OCSPSigning       ///Certificate can be used to sign an Online %Certificate Status Protocol (OCSP) assertion, id = "1.3.6.1.5.5.7.3.9".
  };

  class KeyUsage {
    public:
      KeyUsage() : type_(-1), extended_(false){ };

      KeyUsage(KeyUsageType type);

      /*@param id the type as an identifier string (OID or internal)
        @param section the section this type belongs in
      */
      KeyUsage(const std::string &id, bool extended);

      KeyUsage(const KeyUsage &from) : id_(from.id_), type_(from.type_), extended_(from.extended_) { };

      ~KeyUsage(){ };

      KeyUsage & operator=(const KeyUsage &from) {
        id_ = from.id_;
        type_ = from.type_;
        extended_ = from.extended_;
        return *this;
      };

      bool extended() const { return extended_; };

      KeyUsageType type() const { return (KeyUsageType)type_; };

      std::string id() const { return id_; };

      bool operator<(const KeyUsage &other) const;

      bool operator==(const KeyUsage &other) const;

      inline bool operator!=(const KeyUsage &other) const {
                return !(*this == other);
      }

    private:
        std::string id_;
        int type_;
        bool extended_;
  };
  class KeyUsages {
    private:
      std::list<KeyUsage> keyusages;
      KeyUsage dummy;
    public:
      KeyUsages& operator+= ( const KeyUsages & other ); //the overlap KeyUsage will only be count once
      KeyUsages& operator+= ( const KeyUsage& value );
      KeyUsages& operator= ( const KeyUsages& other ) {
        keyusages.clear();
        for(std::list<KeyUsage>::const_iterator i=other.keyusages.begin(); i!=other.keyusages.end(); i++) keyusages.push_back(*i); return *this;
      }
      const KeyUsage& operator[] ( int num ) const {
        int j = 0;
        for(std::list<KeyUsage>::const_iterator i=keyusages.begin(); i != keyusages.end(); i++, j++) {
          if(j==num) return *i;
        }
        return dummy;
      }
      KeyUsages& Intersection( const KeyUsages & other );
      void Clear() { keyusages.clear(); }
      int Size() { return keyusages.size(); }
      const int Size() const { return keyusages.size(); }
      operator std::string(void) const;
  };

  int check_issued(X509_STORE_CTX* ctx, X509* x, X509* issuer);

  std::string GetOpenSSLError(int err = 0);

  bool parse_extension(X509_EXTENSION* x509_ext, bool& crit, std::string& oid, std::string& ext_val);
  
  bool make_request(X509_REQ* req, const EVP_PKEY* pubkey, const EVP_PKEY* privkey, bool req_attribs, const std::string& conf_file, const std::string& subject_name);

  bool selfsign_request(X509** cert, const X509_REQ* req, const EVP_PKEY* privkey,  const time_t from, const time_t till, const std::string& dn_name, const long serial, const std::string& ext_file);

  bool sign_request(X509** cert, const X509_REQ* req, const EVP_PKEY* issuer_privkey, const X509* issuer_cert, const time_t from, const time_t till, const std::string& dn_name, const std::string& serial_file, const std::string& ext_file);

  bool sign_proxy_request(X509** cert, const X509_REQ* req, const EVP_PKEY* issuer_privkey, const X509* issuer_cert, const time_t from, const time_t till, const std::string& proxypolicy, const std::string& conf_file);

  X509_STORE* setup_verify(const std::string& CAfile, const std::string& CApath);

  //query the OCSP responder to get the certificate status, according to rfc2560
  //this is as the alternative of crl
  bool query_ocsp_responder(const std::string& url, X509* cert_to_check, X509* issuer,
      X509* signercert, EVP_PKEY* signerkey, const std::string& cadir, const std::string& cafile, const std::string& cert_path, std::string& resp_str);

  bool query_ocsp_responder(const std::string& url, const std::string& cert_file, const std::string& issuer_file,
      const std::string& signercert, const std::string& signerkey, const std::string& cadir, const std::string& cafile, std::string& resp_str);

  class OpenSSLInitializer {
   private:
    static pthread_mutex_t * ssl_locks;
    static int ssl_locks_num;
    static unsigned long ssl_id_cb(void);
    static void ssl_locking_cb(int mode, int n, const char * s_, int n_);
   public:
    /// Creating instance of this class will initialize OpenSSL.
    OpenSSLInitializer(void);
    /// It is advisable to call this method upon exiting thread
    /// to avoid memory leaks.
    static void ThreadCleanup(void);
  };

}
}


