
#include <pkcs11-helper-1.0/pkcs11h-core.h>
#include <pkcs11-helper-1.0/pkcs11h-certificate.h>
#include <pkcs11-helper-1.0/pkcs11h-token.h>
#include <pkcs11-helper-1.0/pkcs11h-openssl.h>
#include <string.h>
#include <stdio.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <conio.h>
#endif

#include <openssl/pem.h>

#include "pkcs11util.h"
#include "logger.h"

namespace AuthN {

  Logger& PKCS11UtilLogger = log();

  static PKCS11H_BOOL _pkcs11h_hooks_pin_prompt( void * const global_data,
      void * const user_data, const pkcs11h_token_id_t token,
      const unsigned retry, char * const pin, const size_t pin_max) {
    char prompt[1024];
    char *p = NULL;
    snprintf (prompt, sizeof (prompt), "Please enter '%s' PIN or 'cancel': ", token->display);

#ifndef WIN32
    p = getpass (prompt);
#else
    size_t i = 0;
    char c;
    while (i < pin_max && (c = getch ()) != '\r') {
      pin[i++] = c;
    }
    fprintf (stderr, "\n");
#endif
    strncpy (pin, p, pin_max);
    pin[pin_max-1] = '\0';
    return strcmp (pin, "cancel") != 0;
  }

  bool loadProvider(const std::string& provider) {
    CK_RV rv = pkcs11h_initialize();
    if(rv != CKR_OK) {
      PKCS11UtilLogger.msg(ERROR,"Failed to initiate PKCS#11 engine: %s", pkcs11h_getMessage(rv));
      return false;
    }
    rv = pkcs11h_setPINPromptHook(_pkcs11h_hooks_pin_prompt, NULL); //TODO
    if(rv != CKR_OK) {
      unloadProvider(provider);
      PKCS11UtilLogger.msg(ERROR, "Failed to set PKCS #11 engine pin hook: %s", pkcs11h_getMessage(rv));
     return false;
    }

    rv = pkcs11h_addProvider(provider.c_str(), provider.c_str(), TRUE,
                PKCS11H_PRIVATEMODE_MASK_AUTO, PKCS11H_SLOTEVENT_METHOD_AUTO, 0, TRUE);
    if(rv != CKR_OK) {
      unloadProvider(provider);
      PKCS11UtilLogger.msg(ERROR, "Failed to load driver '%s' for PKCS #11 engine: %s",
                provider.c_str(), pkcs11h_getMessage(rv));
      return false;
    }
    return true;
  }

  void unloadProvider(const std::string& provider) {
    pkcs11h_removeProvider(provider.c_str());
    pkcs11h_terminate();
  }

  bool getCredential(const std::string& certname, X509** x509, RSA** rsa) {
    pkcs11h_certificate_id_list_t issuers;
    pkcs11h_certificate_id_list_t certs;
    pkcs11h_certificate_id_t find = NULL;
    CK_RV rv = pkcs11h_certificate_enumCertificateIds(PKCS11H_ENUM_METHOD_CACHE, NULL,
              PKCS11H_PROMPT_MASK_ALLOW_ALL, &issuers, &certs);
    if(rv != CKR_OK || certs == NULL) {
      PKCS11UtilLogger.msg(ERROR, "Cannot enumerate certificates: %s", pkcs11h_getMessage(rv));
      return false;
    }
    PKCS11UtilLogger.msg(INFO, "Succeed to enumerate certificate");
  
    int i = 0;
    //Try to find the certificate with the same label as given in the input
    for(pkcs11h_certificate_id_list_t cert = certs; cert != NULL; cert = cert->next) {
      std::string label=cert->certificate_id->displayName;
      i++;
      PKCS11UtilLogger.msg(INFO, "The name of the %d certficate is %s \n", i, label.c_str());
      if(certname == label) { 
        pkcs11h_certificate_duplicateCertificateId(&find, cert->certificate_id);
        //TODO: probably it is need to deal with the case that multiple certificate with the same name exists.
        break; 
      }
    }

    pkcs11h_certificate_freeCertificateIdList(issuers);
    pkcs11h_certificate_freeCertificateIdList(certs);

    if(find == NULL) {
      PKCS11UtilLogger.msg(ERROR, "Could not find certificate with the name %s", certname.c_str());
      return false;
    }

    pkcs11h_certificate_t certificate;
    rv = pkcs11h_certificate_create(find, NULL, PKCS11H_PROMPT_MASK_ALLOW_ALL, PKCS11H_PIN_CACHE_INFINITE, &certificate);
    if(rv != CKR_OK) {
      PKCS11UtilLogger.msg(ERROR, "Can not read certificate: %s", pkcs11h_getMessage(rv));
      pkcs11h_certificate_freeCertificateId(find);
      return false;
    }
    pkcs11h_certificate_freeCertificateId(find);

    pkcs11h_openssl_session_t openssl_session = NULL;
    if((openssl_session = pkcs11h_openssl_createSession(certificate)) == NULL) {
      PKCS11UtilLogger.msg(ERROR, "Cannot initialize openssl session to retrieve X509 and RSA");
      pkcs11h_certificate_freeCertificate(certificate);
    }
//////////
    rv = pkcs11h_certificate_ensureKeyAccess(certificate);
    if(rv != CKR_OK) {
      PKCS11UtilLogger.msg(ERROR, "The private key is not accessible: %s", pkcs11h_getMessage(rv));
    }
    else
      PKCS11UtilLogger.msg(INFO, "The private key is accessible");
//////
    certificate = NULL; // the certificate object will be released by openssl_session
 
    bool ret;
    X509* x509_local;
    RSA* rsa_local;
    x509_local = pkcs11h_openssl_session_getX509(openssl_session);
    if(!x509_local) { PKCS11UtilLogger.msg(ERROR, "Cannot get X509 object"); ret = false; }
    rsa_local = pkcs11h_openssl_session_getRSA (openssl_session);
    if(!rsa_local) { PKCS11UtilLogger.msg(ERROR, "Cannot get RSA object"); ret = false; }
    ret = true;

    PKCS11UtilLogger.msg(INFO, "Succeed to get X509 and RSA");
    *x509 = x509_local;
    *rsa = rsa_local;
    pkcs11h_openssl_freeSession (openssl_session);

    EVP_PKEY *pkey;
    pkey = EVP_PKEY_new();
    EVP_PKEY_assign_RSA(pkey, rsa_local);


    std::string str;
    BIO* out = BIO_new(BIO_s_mem());
    const EVP_CIPHER* cipher = EVP_aes_128_cbc();
    std::string passphrase = "secretpw";
    //if(!PEM_write_bio_RSAPrivateKey(out, rsa, cipher, (unsigned char*)(passphrase.c_str()), passphrase.length(), NULL, NULL)) { BIO_free_all(out); return 0; };
    if(!PEM_write_bio_PrivateKey(out,pkey,NULL,NULL,0,NULL,NULL)) { BIO_free_all(out); return 0; };
    //if(!PEM_write_bio_RSAPrivateKey(out, rsa, NULL, NULL, 0, NULL, NULL)) { BIO_free_all(out); return 0; };
    for(;;) {
      char s[256];
      int l = BIO_read(out,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(out);
    std::cout<<"Private key is=== : \n"<<str<<std::endl;


    return ret;
  }

}
